@@ -1,13 +1,14 @@
-import contextvars
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import dataclasses
-import datetime
 import inspect
 import itertools
 import json
 import os
+import platform
 import sys
 import traceback as _traceback
-import urllib.parse
 from types import FrameType, TracebackType
 from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, Iterable, List, MutableSet, NoReturn, Optional, \
     Tuple, Type, TypeVar, Union, cast
@@ -18,6 +19,7 @@ import dramatiq.brokers.stub
 import dramatiq.middleware.age_limit
 import dramatiq.middleware.callbacks
 import dramatiq.middleware.current_message
+import dramatiq.middleware.pipelines
 import dramatiq.middleware.shutdown
 import dramatiq.middleware.time_limit
 import dramatiq.rate_limits.backends
@@ -25,6 +27,7 @@ import dramatiq.rate_limits.concurrent
 import gluetool.log
 import gluetool.sentry
 import gluetool.utils
+import jinja2
 import jinja2.defaults
 import jinja2_ansible_filters.core_filters
 import jsonschema
@@ -33,6 +36,18 @@ import pkg_resources
 import redis
 import ruamel.yaml
 import ruamel.yaml.compat
+import sentry_sdk
+import sentry_sdk.integrations.argv
+import sentry_sdk.integrations.atexit
+import sentry_sdk.integrations.dedupe
+import sentry_sdk.integrations.excepthook
+import sentry_sdk.integrations.logging
+import sentry_sdk.integrations.modules
+import sentry_sdk.integrations.stdlib
+import sentry_sdk.integrations.threading
+import sentry_sdk.serializer
+import sentry_sdk.transport
+import sentry_sdk.utils
 import stackprinter
 from gluetool.result import Error, Ok, Result
 
@@ -49,9 +64,13 @@ jinja2.defaults.DEFAULT_FILTERS.update(
 # Now we can import our stuff without any fear we'd miss DEFAULT_FILTERS update
 from . import db as artemis_db  # noqa: E402
 from . import middleware as artemis_middleware  # noqa: E402
+from .knobs import Knob  # noqa: E402
+from .knobs import KNOB_DEPLOYMENT_ENVIRONMENT, KNOB_LOGGING_SENTRY, KNOB_SENTRY_BASE_URL, \
+    KNOB_SENTRY_DISABLE_CERT_VERIFICATION, KNOB_SENTRY_DSN  # noqa: E402
 
 if TYPE_CHECKING:
     from .environment import Environment
+    from .tasks import TaskCall
 
 stackprinter.set_excepthook(
     style='darkbg2',
@@ -67,6 +86,9 @@ stackprinter.set_excepthook(
 #: for this value - if you want to use different port, do specify it in ``BROKER_URL`` then.
 DEFAULT_RABBITMQ_PORT = 5672
 
+#: Default date/time format.
+DATETIME_FMT: str = '%Y-%m-%dT%H:%M:%S.%f'
+
 
 ExceptionInfoType = Union[
     # returned by sys.exc_info()
@@ -100,11 +122,6 @@ FailureDetailsType = Dict[str, Any]
 JSONType = Union[str, int, float, List[Any], Dict[Any, Any], None]
 
 
-# Special context variable for YAML processor. `ruamel.YAML` instances keep internal state, and as such they cannot
-# be shared between threads without some kind of serialization. Rather than introducing a lock, we can use context
-# and keep a `YAML` instance for each thread. With a wrapper function, we can initialize missing instances when
-# accessed for the first time (`contextvars` package does not support default factory).
-_YAML: contextvars.ContextVar[Optional[ruamel.yaml.main.YAML]] = contextvars.ContextVar('_YAML', default=None)
 _YAML_DUMPABLE_CLASSES: MutableSet[Type[object]] = set()
 
 
@@ -113,20 +130,90 @@ def get_yaml() -> ruamel.yaml.main.YAML:
     Return a fully initialized instance of YAML processor.
     """
 
-    YAML = _YAML.get()
+    YAML = gluetool.utils.YAML()
 
-    if YAML is None:
-        YAML = gluetool.utils.YAML()
-        _YAML.set(YAML)
-
-        for cls in _YAML_DUMPABLE_CLASSES:
-            YAML.register_class(cls)
+    for cls in _YAML_DUMPABLE_CLASSES:
+        YAML.register_class(cls)
 
     return YAML
 
 
-# Gluetool Sentry instance
-gluetool_sentry = gluetool.sentry.Sentry()
+# This knob needs to live in this module, because its default value includes
+# __VERSION__ we can't import from knobs module (circular import).
+def cast_release(release_template: str) -> str:
+    """
+    Apply templating to user-defined release string.
+
+    Raw "release" may refer to ``__VERSION__``, render it as a template.
+    """
+
+    return cast(str, jinja2.Template(release_template).render(__VERSION__=__VERSION__))
+
+
+KNOB_RELEASE: Knob[str] = Knob(
+    'deployment.release',
+    'Optional name of the Artemis release (e.g. "0.0.35", "artemis@v0.0.35", "artemis@{{ __VERSION__ }}, etc.).',
+    has_db=False,
+    envvar='ARTEMIS_RELEASE',
+    cast_from_str=cast_release,
+    default=__VERSION__
+)
+
+
+class Sentry:
+    def __init__(self) -> None:
+        self.enabled = False
+
+        if KNOB_SENTRY_DSN.value in (None, 'undefined'):
+            return
+
+        self.enabled = True
+
+        if KNOB_SENTRY_DISABLE_CERT_VERIFICATION.value is True:
+            def _get_pool_options(
+                self: sentry_sdk.transport.HttpTransport,
+                ca_certs: Any
+            ) -> Dict[str, Any]:
+                return {
+                    # num_pools is a bit cryptic, btu comes from the original method
+                    'num_pools': 2,
+                    'cert_reqs': 'CERT_NONE'
+                }
+
+            sentry_sdk.transport.HttpTransport._get_pool_options = _get_pool_options  # type: ignore[assignment]
+
+        # Controls how many variables and other items are captured in event and stack frames. The default
+        # value of 10 is pretty small, 1000 should be more than enough for anything we ever encounter.
+        sentry_sdk.serializer.MAX_DATABAG_BREADTH = 1000
+
+        sentry_sdk.init(
+            dsn=KNOB_SENTRY_DSN.value,
+            release=KNOB_RELEASE.value,
+            environment=KNOB_DEPLOYMENT_ENVIRONMENT.value,
+            server_name=platform.node(),
+            debug=KNOB_LOGGING_SENTRY.value,
+            # log all issues - if we ever decide we need less, we can add knobs to control these
+            sample_rate=1.0,
+            traces_sample_rate=1.0,
+            # We need to override one parameter of on of the default integrations,
+            # so we're doomed to list all of them.
+            integrations=[
+                # Disable sending any log messages as standalone events
+                sentry_sdk.integrations.logging.LoggingIntegration(event_level=None),
+                # The rest is just default list of integrations.
+                # https://docs.sentry.io/platforms/python/configuration/integrations/default-integrations/
+                sentry_sdk.integrations.stdlib.StdlibIntegration(),
+                sentry_sdk.integrations.excepthook.ExcepthookIntegration(),
+                sentry_sdk.integrations.dedupe.DedupeIntegration(),
+                sentry_sdk.integrations.atexit.AtexitIntegration(),
+                sentry_sdk.integrations.modules.ModulesIntegration(),
+                sentry_sdk.integrations.argv.ArgvIntegration(),
+                sentry_sdk.integrations.threading.ThreadingIntegration()
+            ]
+        )
+
+
+SENTRY = Sentry()
 
 
 class SerializableContainer:
@@ -135,10 +222,10 @@ class SerializableContainer:
     """
 
     def __init__(self, *args: Any, **kwargs: Any) -> None:
-        super(SerializableContainer, self).__init__(*args, **kwargs)  # type: ignore
+        super().__init__(*args, **kwargs)
 
     # All classes derived from SerializableContainer can be represented as YAML, because they
-    # inherit the `to_yaml()` method, which then depends on `serialize_to_json()` - and what
+    # inherit the `to_yaml()` method, which then depends on `serialize()` - and what
     # we can represent as JSON, we can for sure represent as YAML as well.
     #
     # Instead of using decorator to mark classes derived from this one, we let interpreter call
@@ -148,11 +235,38 @@ class SerializableContainer:
         Register given subclass as capable of being represented as YAML.
         """
 
-        super(SerializableContainer, cls).__init_subclass__()
+        super().__init_subclass__()
 
         _YAML_DUMPABLE_CLASSES.add(cls)
 
-    def serialize_to_json(self) -> Dict[str, Any]:
+    def __str__(self) -> str:
+        """
+        Return text representation of the container.
+
+        :returns: human-readable rendering of the container.
+        """
+
+        return self.serialize_to_yaml()
+
+    def __repr__(self) -> str:
+        """
+        Return text representation of the container.
+
+        :returns: human-readable rendering of the container.
+        """
+
+        return self.serialize_to_yaml()
+
+    def serialize(self) -> Dict[str, Any]:
+        """
+        Return Python built-in types representing the content of this container.
+
+        Works in a recursive manner, every container member that's a subclass of :py:class:`SerializableContainer`
+        is processed as well.
+
+        See :py:meth:`unserialize` for the reversal operation.
+        """
+
         serialized = dataclasses.asdict(self)
 
         for field in dataclasses.fields(self):
@@ -162,12 +276,20 @@ class SerializableContainer:
             if not issubclass(field.type, SerializableContainer):
                 continue
 
-            serialized[field.name] = getattr(self, field.name).serialize_to_json()
+            serialized[field.name] = getattr(self, field.name).serialize()
 
         return serialized
 
     @classmethod
-    def unserialize_from_json(cls: Type[S], serialized: Dict[str, Any]) -> S:
+    def unserialize(cls: Type[S], serialized: Dict[str, Any]) -> S:
+        """
+        Create container instance representing the content described with Python built-in types.
+
+        Every container member whose type is a subclass of :py:class:`SerializableContainer` is restored as well.
+
+        See :py:meth:`serialize` for the reversal operation.
+        """
+
         unserialized = cls(**serialized)
 
         for field in dataclasses.fields(unserialized):
@@ -180,20 +302,61 @@ class SerializableContainer:
             if field.name not in serialized:
                 continue
 
-            setattr(unserialized, field.name, field.type.unserialize_from_json(serialized[field.name]))
+            setattr(unserialized, field.name, field.type.unserialize(serialized[field.name]))
 
         return unserialized
 
-    def serialize_to_str(self) -> str:
-        return json.dumps(self.serialize_to_json())
+    def serialize_to_json(self) -> str:
+        """
+        Return JSON blob representing the content of this container.
+
+        Works in a recursive manner, every container member that's a subclass of :py:class:`SerializableContainer`
+        is processed as well.
+
+        See :py:meth:`unserialize_from_json` for the reversal operation.
+        """
+
+        return json.dumps(self.serialize())
 
     @classmethod
-    def unserialize_from_str(cls: Type[S], serialized: str) -> S:
-        return cls.unserialize_from_json(json.loads(serialized))
+    def unserialize_from_json(cls: Type[S], serialized: str) -> S:
+        """
+        Create container instance representing the content described with a JSON blob.
+
+        Every container member whose type is a subclass of :py:class:`SerializableContainer` is restored as well.
+
+        See :py:meth:`serialize_to_json` for the reversal operation.
+        """
+
+        return cls.unserialize(json.loads(serialized))
+
+    def serialize_to_yaml(self) -> str:
+        """
+        Return YAML blob representing the content of this container.
+
+        Works in a recursive manner, every container member that's a subclass of :py:class:`SerializableContainer`
+        is processed as well.
+
+        See :py:meth:`unserialize_from_yaml` for the reversal operation.
+        """
+
+        return format_dict_yaml(self.serialize())
+
+    @classmethod
+    def unserialize_from_yaml(cls: Type[S], serialized: str) -> S:
+        """
+        Create container instance representing the content described with a YAML blob.
+
+        Every container member whose type is a subclass of :py:class:`SerializableContainer` is restored as well.
+
+        See :py:meth:`serialize_to_yaml` for the reversal operation.
+        """
+
+        return cls.unserialize(get_yaml().load(serialized))
 
     @classmethod
     def to_yaml(cls, representer: ruamel.yaml.representer.Representer, container: S) -> Any:
-        return representer.represent_dict(container.serialize_to_json())
+        return representer.represent_dict(container.serialize())
 
 
 # Two logging helpers, very similar to `format_dict` and `log_dict`, but emitting a YAML-ish output.
@@ -208,7 +371,13 @@ def format_dict_yaml(data: Any) -> str:
 
     ruamel.yaml.scalarstring.walk_tree(data)
 
-    YAML.dump(data, stream)
+    def strip_document_end_marker(s: str) -> str:
+        if s.endswith('...\n'):
+            s = s[:-4]
+
+        return s.strip()
+
+    YAML.dump(data, stream, transform=strip_document_end_marker)
 
     return stream.getvalue()
 
@@ -245,6 +414,9 @@ def process_output_to_str(output: gluetool.utils.ProcessOutput, stream: str = 's
     return cast(str, stream_content.decode('utf-8'))
 
 
+_DEFAULT_FAILURE_LOG_LABEL = 'failure'
+
+
 class Failure:
     """
     Bundles exception related info.
@@ -270,6 +442,7 @@ class Failure:
         scrubbed_command: Optional[List[str]] = None,
         command_output: Optional[gluetool.utils.ProcessOutput] = None,
         environment: Optional['Environment'] = None,
+        task_call: Optional['TaskCall'] = None,
         **details: Any
     ):
         self.message = message
@@ -302,6 +475,9 @@ class Failure:
         if environment:
             self.details['environment'] = environment
 
+        if task_call:
+            self.details['task_call'] = task_call
+
         if exc_info:
             self.exception = exc_info[1]
 
@@ -348,10 +524,9 @@ class Failure:
         scrubbed_command: Optional[List[str]] = None,
         command_output: Optional[gluetool.utils.ProcessOutput] = None,
         environment: Optional['Environment'] = None,
+        task_call: Optional['TaskCall'] = None,
         **details: Any
-    ):
-        # type: (...) -> Failure
-
+    ) -> 'Failure':
         return Failure(
             message,
             exc_info=(
@@ -365,6 +540,7 @@ class Failure:
             scrubbed_command=scrubbed_command,
             command_output=command_output,
             environment=environment,
+            task_call=task_call,
             **details
         )
 
@@ -378,6 +554,7 @@ class Failure:
         scrubbed_command: Optional[List[str]] = None,
         command_output: Optional[gluetool.utils.ProcessOutput] = None,
         environment: Optional['Environment'] = None,
+        task_call: Optional['TaskCall'] = None,
         **details: Any
     ) -> 'Failure':
         """
@@ -404,6 +581,7 @@ class Failure:
             scrubbed_command=scrubbed_command,
             command_output=command_output,
             environment=environment,
+            task_call=task_call,
             **details
         )
 
@@ -413,6 +591,7 @@ class Failure:
         scrubbed_command: Optional[List[str]] = None,
         command_output: Optional[gluetool.utils.ProcessOutput] = None,
         environment: Optional['Environment'] = None,
+        task_call: Optional['TaskCall'] = None,
         **details: Any
     ) -> 'Failure':
         self.details.update(details)
@@ -426,6 +605,9 @@ class Failure:
         if environment:
             self.details['environment'] = environment
 
+        if task_call:
+            self.details['task_call'] = task_call
+
         return self
 
     @classmethod
@@ -457,6 +639,10 @@ class Failure:
             'type': str(type(exc))
         }
 
+    @property
+    def command_output(self) -> Optional[gluetool.utils.ProcessOutput]:
+        return self.details.get('command_output')
+
     def get_event_details(self) -> Dict[str, Any]:
         """
         Returns a mapping of failure details, suitable for storing in DB as a guest event details.
@@ -479,7 +665,10 @@ class Failure:
         event_details.pop('job_results', None)
 
         if 'environment' in event_details:
-            event_details['environment'] = event_details['environment'].serialize_to_json()
+            event_details['environment'] = event_details['environment'].serialize()
+
+        if 'task_call' in event_details:
+            event_details['task_call'] = event_details['task_call'].serialize()
 
         if self.caused_by:
             event_details['caused_by'] = self.caused_by.get_event_details()
@@ -493,14 +682,39 @@ class Failure:
         return event_details
 
     @classmethod
-    def _get_sentry_stack_info(cls, frames: _traceback.StackSummary) -> Dict[str, Any]:
+    def _serialize_traceback(cls, message: str, frames: _traceback.StackSummary, ) -> Dict[str, Any]:
         """
-        Based on Raven's ``get_stack_info``. Because Raven is quite outdated, we need to convert
-        the traceback to a format Sentry accepts. The original method cannot deal with ``FrameSummary``
-        objects used by Python 3.5+, it understands only the old frame objects, witg ``f_*`` attributes.
+        Based on Sentry's stack trace serialization, this helper takes care of serializing
+        a traceback saved by ``Failure`` instance itself, i.e. when there was no exception.
         """
 
-        from raven.utils.stacks import get_lines_from_file, slim_frame_data
+        # Convert weird string-ish objects into strings we can use as part of serialized frames.
+        # Sentry SDK uses several types to carry lines of code, deal with all of them.
+        def _stringify(
+            v: Union[
+                str,
+                sentry_sdk.utils.AnnotatedValue,
+                List[str],
+                List[Union[sentry_sdk.utils.AnnotatedValue, str]],
+                None
+            ]
+        ) -> str:
+            if v is None:
+                return ''
+
+            if isinstance(v, str):
+                return v
+
+            if isinstance(v, sentry_sdk.utils.AnnotatedValue):
+                return str(v.value)
+
+            if isinstance(v, list):
+                return '\n'.join([
+                    (s.value or '') if isinstance(s, sentry_sdk.utils.AnnotatedValue) else s
+                    for s in v
+                ])
+
+            return str(v)
 
         result = []
         for frame in frames:
@@ -520,12 +734,12 @@ class Failure:
             if line is not None:
                 # Lines are indexed from 0, but human representation starts with 1: "1st line".
                 # Hence the decrement.
-                pre_context, context_line, post_context = get_lines_from_file(filename, lineno - 1, 5)
+                pre_context, context_line, post_context = sentry_sdk.utils.get_lines_from_file(filename, lineno - 1)
 
                 frame_result.update({
-                    'pre_context': pre_context,
-                    'context_line': context_line,
-                    'post_context': post_context,
+                    'pre_context': _stringify(pre_context),
+                    'context_line': _stringify(context_line),
+                    'post_context': _stringify(post_context)
                 })
 
             if frame.locals:
@@ -533,24 +747,40 @@ class Failure:
 
             result.append(frame_result)
 
-        stackinfo = {
-            'frames': slim_frame_data(result),
+        return {
+            # "Magic" structures, see sentry_sdk.utils for implementation creating these for exceptions.
+            'module': None,
+            # Type and value are required - but we have none of them, obviously, otherwise we'd use
+            # the exception and its type.
+            #
+            # Instead of an actual type, to comply with the way Sentry displays events with a traceback,
+            # we store the failure message as `type`, and ignore the `value` field. This should present
+            # nice and good looking event.
+            'type': message,
+            'value': None,
+            'mechanism': {
+                'type': 'generic'
+            },
+            'stacktrace': {
+                'frames': result
+            }
         }
 
-        return stackinfo
-
     def get_sentry_details(self) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
         """
         Returns three mappings, data, tags and extra, accepted by Sentry as issue details.
         """
 
-        data: Dict[str, Any] = {}
+        from .knobs import KNOB_DEPLOYMENT, KNOB_DEPLOYMENT_ENVIRONMENT
+
+        event: Dict[str, Any] = {}
         tags: Dict[str, str] = {}
         extra: Dict[str, Any] = self.details.copy()
 
         extra['message'] = self.message
         extra['recoverable'] = self.recoverable
         extra['fail_guest_request'] = self.fail_guest_request
+        extra['env'] = os.environ.copy()
 
         if 'scrubbed_command' in extra:
             extra['scrubbed_command'] = gluetool.utils.format_command_line([extra['scrubbed_command']])
@@ -570,13 +800,41 @@ class Failure:
         if 'poolname' in self.details:
             tags['poolname'] = self.details['poolname']
 
-        if self.traceback:
+        if self.exc_info:
+            event, _ = sentry_sdk.utils.event_from_exception(self.exc_info)
+
+        elif self.traceback:
             # Convert our traceback to format understood by Sentry, and store it in `data['stacktrace']` where Sentry
             # expects it to find when generating the message for submission.
-            data['stacktrace'] = Failure._get_sentry_stack_info(self.traceback)
+
+            event.update({
+                'level': 'error',
+                'exception': {
+                    'values': [
+                        Failure._serialize_traceback(self.message, self.traceback)
+                    ]
+                }
+            })
 
         if 'environment' in extra:
-            extra['environment'] = extra['environment'].serialize_to_json()
+            extra['environment'] = extra['environment'].serialize()
+
+        if 'task_call' in extra:
+            extra['task_call'] = extra['task_call'].serialize()
+
+        # Special tag, "server_name", is used by Sentry for tracking issues per servers.
+        tags['server_name'] = platform.node()
+
+        if KNOB_DEPLOYMENT.value:
+            tags['deployment'] = KNOB_DEPLOYMENT.value
+
+        # Special tag, "environment", is used by Sentry for tracking issues per environment.
+        if KNOB_DEPLOYMENT_ENVIRONMENT.value:
+            tags['environment'] = KNOB_DEPLOYMENT_ENVIRONMENT.value
+
+        event.update({
+            'message': f'Failure: {self.message}'
+        })
 
         tags.update({
             key: value
@@ -593,7 +851,7 @@ class Failure:
                 'extra': caused_by_extra
             }
 
-        return data, tags, extra
+        return event, tags, extra
 
     def get_log_details(self) -> Dict[str, Any]:
         """
@@ -610,7 +868,10 @@ class Failure:
             details['exception'] = self._exception_details(self.exception, self.details.get('scrubbed_command'))
 
         if self.exc_info:
-            details['traceback'] = stackprinter.format(self.exc_info)  # noqa
+            details['traceback'] = '\n'.join(
+                line.rstrip()
+                for line in stackprinter.format(self.exc_info, line_wrap=False).splitlines()  # noqa: FS002
+            )
 
         if 'scrubbed_command' in details:
             details['scrubbed_command'] = gluetool.utils.format_command_line([details['scrubbed_command']])
@@ -624,7 +885,10 @@ class Failure:
             }
 
         if 'environment' in details:
-            details['environment'] = details['environment'].serialize_to_json()
+            details['environment'] = details['environment'].serialize()
+
+        if 'task_call' in details:
+            details['task_call'] = details['task_call'].serialize()
 
         if self.caused_by:
             details['caused-by'] = self.caused_by.get_log_details()
@@ -637,46 +901,57 @@ class Failure:
 
         return details
 
+    def _printable(
+        self,
+        label: str = _DEFAULT_FAILURE_LOG_LABEL
+    ) -> str:
+        return f'{label}\n\n{format_dict_yaml(self.get_log_details())}'
+
+    def __str__(self) -> str:
+        return self._printable()
+
+    def __repr__(self) -> str:
+        return f'<Failure: message="{self.message}">'
+
     def log(
         self,
         log_fn: gluetool.log.LoggingFunctionType,
-        label: Optional[str] = None
+        label: str = _DEFAULT_FAILURE_LOG_LABEL
     ) -> None:
         exc_info = self.exc_info if self.exc_info else (None, None, None)
 
-        details = self.get_log_details()
-
-        if not label:
-            label = 'failure'
-
-        log_fn(f'{label}\n\n{format_dict_yaml(details)}', exc_info=exc_info)
+        log_fn(self._printable(label=label), exc_info=exc_info)
 
     def submit_to_sentry(self, logger: gluetool.log.ContextAdapter, **additional_tags: Any) -> None:
         if self.submited_to_sentry:
             return
 
-        data, tags, extra = self.get_sentry_details()
+        if not SENTRY.enabled:
+            return
+
+        event, tags, extra = self.get_sentry_details()
 
         if additional_tags:
             tags.update(additional_tags)
 
-        try:
-            self.sentry_event_id = gluetool_sentry.submit_message(
-                f'Failure: {self.message}',
-                exc_info=self.exc_info,
-                data=data,
-                tags=tags,
-                extra=extra
-            )
+        with sentry_sdk.push_scope() as scope:
+            for name, value in tags.items():
+                scope.set_tag(name, value)
 
-        except Exception as exc:
-            Failure.from_exc('failed to submit to Sentry', exc).handle(logger, sentry=False)
+            for name, value in extra.items():
+                scope.set_extra(name, value)
 
-        else:
-            self.submited_to_sentry = True
+            try:
+                self.sentry_event_id = sentry_sdk.capture_event(event, scope=scope)
 
-        if self.sentry_event_id:
-            self.sentry_event_url = gluetool_sentry.event_url(self.sentry_event_id, logger=logger)
+            except Exception as exc:
+                Failure.from_exc('failed to submit to Sentry', exc).handle(logger, sentry=False)
+
+            else:
+                self.submited_to_sentry = True
+
+            if self.sentry_event_id and KNOB_SENTRY_BASE_URL.value not in (None, 'undefined'):
+                self.sentry_event_url = f'{KNOB_SENTRY_BASE_URL.value}/?query={self.sentry_event_id}'
 
     def reraise(self) -> NoReturn:
         if self.exception:
@@ -687,7 +962,7 @@ class Failure:
     def handle(
         self,
         logger: gluetool.log.ContextAdapter,
-        label: Optional[str] = None,
+        label: str = _DEFAULT_FAILURE_LOG_LABEL,
         sentry: bool = True,
         **details: Any
     ) -> None:
@@ -706,8 +981,7 @@ def get_logger() -> gluetool.log.ContextAdapter:
 
     return gluetool.log.Logging.setup_logger(
         level=KNOB_LOGGING_LEVEL.value,
-        json_output=KNOB_LOGGING_JSON.value,
-        sentry=gluetool_sentry
+        json_output=KNOB_LOGGING_JSON.value
     )
 
 
@@ -723,56 +997,93 @@ def get_config() -> Dict[str, Any]:
     )
 
 
-def get_broker() -> dramatiq.brokers.rabbitmq.RabbitmqBroker:
-    if os.getenv('IN_TEST', None):
-        broker = dramatiq.brokers.stub.StubBroker(middleware=[
-            dramatiq.middleware.age_limit.AgeLimit(),
-            dramatiq.middleware.time_limit.TimeLimit(),
-            dramatiq.middleware.shutdown.ShutdownNotifications(notify_shutdown=True),
-            dramatiq.middleware.callbacks.Callbacks(),
-            dramatiq.middleware.GroupCallbacks(dramatiq.rate_limits.backends.stub.StubBackend()),
-            dramatiq.middleware.current_message.CurrentMessage(),
-            artemis_middleware.Prometheus(),
-            artemis_middleware.Retries(),
-            periodiq.PeriodiqMiddleware()
-        ])
+def get_broker_middleware(logger: gluetool.log.ContextAdapter) -> List[dramatiq.Middleware]:
+    from .knobs import KNOB_WORKER_PROCESS_METRICS_ENABLED, KNOB_WORKER_TRAFFIC_METRICS_ENABLED
 
-    else:
-        # We need a better control over some aspects of our broker connection, e.g. heartbeat
-        # and timeouts. This is possible, but we cannt use broker URL as a argument, this is
-        # not supported with Dramatiq and Pika. Therefore, we need to parse the URL and construct
-        # connection parameters, with expected content, and add more details when needed.
+    middleware: List[dramatiq.Middleware] = []
 
-        import pika
+    if KNOB_WORKER_PROCESS_METRICS_ENABLED.value is True:
+        from .knobs import KNOB_WORKER_PROCESS_METRICS_UPDATE_TICK
 
-        from .knobs import KNOB_BROKER_CONFIRM_DELIVERY, KNOB_BROKER_HEARTBEAT_TIMEOUT, KNOB_BROKER_URL
+        middleware.append(
+            artemis_middleware.WorkerMetrics(
+                f'worker-{platform.node()}-{os.getpid()}',
+                KNOB_WORKER_PROCESS_METRICS_UPDATE_TICK.value
+            )
+        )
 
-        # TODO: for actual limiter, we would need to throw in either Redis or Memcached.
-        # Using stub and a dummy key for now, but it's just not going to do its job properly.
+    if KNOB_WORKER_TRAFFIC_METRICS_ENABLED.value is True:
+        middleware.append(
+            artemis_middleware.WorkerTraffic(
+                logger,
+                get_cache(logger),
+                f'worker-{platform.node()}-{os.getpid()}'
+            )
+        )
+
+    middleware += [
+        dramatiq.middleware.age_limit.AgeLimit(),
+        dramatiq.middleware.time_limit.TimeLimit(),
+        dramatiq.middleware.shutdown.ShutdownNotifications(notify_shutdown=True),
+        dramatiq.middleware.callbacks.Callbacks(),
+        dramatiq.middleware.GroupCallbacks(dramatiq.rate_limits.backends.stub.StubBackend()),
+        dramatiq.middleware.pipelines.Pipelines(),
+        artemis_middleware.CurrentMessage(),
+        artemis_middleware.Prometheus(),
+        artemis_middleware.Retries(),
+        periodiq.PeriodiqMiddleware()
+    ]
+
+    return middleware
+
+
+def get_broker(
+    logger: gluetool.log.ContextAdapter,
+    application_name: Optional[str] = None
+) -> dramatiq.brokers.rabbitmq.RabbitmqBroker:
+    from .knobs import KNOB_BROKER_CONFIRM_DELIVERY, KNOB_BROKER_URL
+
+    middleware: List[dramatiq.Middleware] = get_broker_middleware(logger)
+
+    # TODO: for actual limiter, we would need to throw in either Redis or Memcached.
+    # Using stub and a dummy key for now, but it's just not going to do its job properly.
+
+    broker_url = KNOB_BROKER_URL.value
+
+    # Client properties must be encoded into URL, Pika does not allow `url` + `client_properties` at the same time.
+    client_properties: Dict[str, str] = {}
+
+    if application_name is not None:
+        client_properties['connection_name'] = application_name
+
+    if client_properties:
+        import urllib.parse
 
         parsed_url = urllib.parse.urlparse(KNOB_BROKER_URL.value)
 
-        broker = dramatiq.brokers.rabbitmq.RabbitmqBroker(
-            confirm_delivery=KNOB_BROKER_CONFIRM_DELIVERY.value,
-            middleware=[
-                dramatiq.middleware.age_limit.AgeLimit(),
-                dramatiq.middleware.time_limit.TimeLimit(),
-                dramatiq.middleware.shutdown.ShutdownNotifications(notify_shutdown=True),
-                dramatiq.middleware.callbacks.Callbacks(),
-                dramatiq.middleware.GroupCallbacks(dramatiq.rate_limits.backends.stub.StubBackend()),
-                dramatiq.middleware.current_message.CurrentMessage(),
-                artemis_middleware.Prometheus(),
-                artemis_middleware.Retries(),
-                periodiq.PeriodiqMiddleware()
-            ],
-            parameters=[{
-                'host': parsed_url.hostname,
-                'port': int(parsed_url.port or DEFAULT_RABBITMQ_PORT),
-                'credentials': pika.PlainCredentials(parsed_url.username, parsed_url.password),
-                'heartbeat': KNOB_BROKER_HEARTBEAT_TIMEOUT.value,
-                'blocked_connection_timeout': KNOB_BROKER_HEARTBEAT_TIMEOUT.value
-            }]
-        )
+        parsed_query: Dict[str, Union[str, Dict[str, str]]] = {
+            k: v
+            for k, v in urllib.parse.parse_qsl(parsed_url.query)
+        }
+
+        parsed_query['client_properties'] = client_properties
+
+        broker_url = urllib.parse.ParseResult(
+            scheme=parsed_url.scheme,
+            netloc=parsed_url.netloc,
+            path=parsed_url.path,
+            params=parsed_url.params,
+            query=urllib.parse.urlencode(parsed_query),
+            fragment=parsed_url.fragment
+        ).geturl()
+
+    logger.debug(f'final broker URL is {broker_url}')
+
+    broker = dramatiq.brokers.rabbitmq.RabbitmqBroker(
+        confirm_delivery=KNOB_BROKER_CONFIRM_DELIVERY.value,
+        middleware=middleware,
+        url=broker_url
+    )
 
     dramatiq.set_broker(broker)
 
@@ -822,161 +1133,63 @@ def safe_call(fn: Callable[..., T], *args: Any, **kwargs: Any) -> Result[T, Fail
         return Error(Failure.from_exc('exception raised inside a safe block', exc))
 
 
+# Once mypy implements support for PEP 612, something like this would be the way to go.
+# https://github.com/python/mypy/issues/8645
 #
-# Helpful cache primitives
+# P = ParamSpec('P')
 #
-def refresh_cached_set(
-    cache: redis.Redis,
-    key: str,
-    items: Dict[str, SerializableContainer]
-) -> Result[None, Failure]:
+# def handle_failure(default: Union[T, None] = None) -> Callable[Concatenate[gluetool.log.ContextAdapter, P], T]:
+#    def decorator(fn: Callable[P, T]) -> Callable[Concatenate[gluetool.log.ContextAdapter, P], T]:
+#        @functools.wraps(fn)
+#        def wrapper(logger: gluetool.log.ContextAdapter, *args: Any, **kwargs: Any) -> T:
+#            try:
+#                return fn(*args, **kwargs)
+#
+#            except Exception as exc:
+#                Failure.from_exc('exception raised inside a safe block', exc).handle(logger)
+#
+#                return default
+#
+#        return wrapper
+#
+#    return decorator
+
+
+def safe_call_and_handle(
+    logger: gluetool.log.ContextAdapter,
+    fn: Callable[..., T],
+    *args: Any,
+    **kwargs: Any
+) -> Optional[T]:
     """
-    Refresh a cache entry with a given set of items.
+    Call given function, with provided arguments. If the call fails, log the resulting failure before returning it.
 
-    The cache stores items serialized as JSON blobs, reachable by the key given to them by the set.
+    .. note::
 
-    :param key: cache key storing the set.
-    :param items: set of items to store.
+       Similar to :py:func:`tft.artemis.safe_call`, but
+
+       * does handle the potential failure, and
+       * does not return :py:class:`Result` instance but either the bare value or ``None``.
+
+       This is on purpose, because such a helper fits the needs of cache-related helpers we use for tracking
+       metrics. The failure to communicate with the case isn't a reason to interrupt the main body of work,
+       but it still needs to be reported.
+
+    :param logger: logger to use for logging.
+    :param fn: function to decorate.
+    :param args: positional arguments of ``fn``.
+    :param kwargs: keyword arguments of ``fn``.
+    :returns: if an exception was raised during the function call, a failure is logged and ``safe_call_and_handle``
+        returns ``None``. Otherwise, the return value of the call is returned.
     """
 
-    key_updated = f'{key}.updated'
+    try:
+        return fn(*args, **kwargs)
 
-    if not items:
-        # When we get an empty set of items, we should remove the key entirely, to make queries looking for
-        # return `None` aka "not found". It's the same as if we'd try to remove all entries, just with one
-        # action.
-        safe_call(
-            cast(Callable[[str], None], cache.delete),
-            key
-        )
+    except Exception as exc:
+        Failure.from_exc('exception raised inside a safe block', exc).handle(logger)
 
-        safe_call(
-            cast(Callable[[str, float], None], cache.set),
-            key_updated,
-            datetime.datetime.timestamp(datetime.datetime.utcnow())
-        )
-
-        return Ok(None)
-
-    # Two steps: create new structure, and replace the old one. We cannot check the old one
-    # and remove entries that are no longer valid.
-    new_key = f'{key}.new'
-
-    r_action = safe_call(
-        cast(Callable[[str, str, Dict[str, str]], None], cache.hmset),
-        new_key,
-        {
-            item_key: json.dumps(item.serialize_to_json())
-            for item_key, item in items.items()
-        }
-    )
-
-    if r_action.is_error:
-        return Error(r_action.unwrap_error())
-
-    safe_call(
-        cast(Callable[[str, str], None], cache.rename),
-        new_key,
-        key
-    )
-
-    safe_call(
-        cast(Callable[[str, float], None], cache.set),
-        key_updated,
-        datetime.datetime.timestamp(datetime.datetime.utcnow())
-    )
-
-    return Ok(None)
-
-
-def get_cached_items(
-    cache: redis.Redis,
-    key: str,
-    item_klass: Type[S]
-) -> Result[Optional[Dict[str, S]], Failure]:
-    """
-    Return cached items of a given type in the form of a mapping between their names and the instances.
-
-    Serves as a helper function for fetching homogenous sets of objects, like pool image infos.
-
-    See :py:func:`get_cached_items_as_list` for the variant returning items in a list.
-    """
-
-    r_fetch = safe_call(
-        cast(Callable[[str], Optional[Dict[bytes, bytes]]], cache.hgetall),
-        key,
-    )
-
-    if r_fetch.is_error:
-        return Error(r_fetch.unwrap_error())
-
-    serialized = r_fetch.unwrap()
-
-    if serialized is None:
-        return Ok(None)
-
-    items: Dict[str, S] = {}
-
-    for item_key, item_serialized in serialized.items():
-        r_unserialize = safe_call(item_klass.unserialize_from_json, json.loads(item_serialized.decode('utf-8')))
-
-        if r_unserialize.is_error:
-            return Error(r_unserialize.unwrap_error())
-
-        items[item_key.decode('utf-8')] = r_unserialize.unwrap()
-
-    return Ok(items)
-
-
-def get_cached_items_as_list(
-    cache: redis.Redis,
-    key: str,
-    item_klass: Type[S]
-) -> Result[List[S], Failure]:
-    """
-    Return cached items of a given type in the form of a list of instances.
-
-    Serves as a helper function for fetching homogenous sets of objects, like pool image infos.
-
-    See :py:func:`get_cached_items` for the variant returning items in the form of a mapping.
-    """
-
-    r_fetch = get_cached_items(cache, key, item_klass)
-
-    if r_fetch.is_error:
-        return Error(r_fetch.unwrap_error())
-
-    items = r_fetch.unwrap()
-
-    return Ok(list(items.values()) if items else [])
-
-
-def get_cached_item(
-    cache: redis.Redis,
-    key: str,
-    item_key: str,
-    item_klass: Type[S]
-) -> Result[Optional[S], Failure]:
-    r_fetch = safe_call(
-        cast(Callable[[str, str], Optional[bytes]], cache.hget),
-        key,
-        item_key
-    )
-
-    if r_fetch.is_error:
-        return Error(r_fetch.unwrap_error())
-
-    serialized = r_fetch.unwrap()
-
-    if serialized is None:
-        return Ok(None)
-
-    r_unserialize = safe_call(item_klass.unserialize_from_json, json.loads(serialized.decode('utf-8')))
-
-    if r_unserialize.is_error:
-        return Error(r_unserialize.unwrap_error())
-
-    return Ok(r_unserialize.unwrap())
+        return None
 
 
 #: Custom type for JSON schema. We don't expect the schema structure though, all we do is loading it
