@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import base64
 import dataclasses
 import json
@@ -15,7 +18,7 @@ from molten.errors import HTTPError
 from molten.http.status_codes import HTTP_200
 
 from .. import Failure
-from ..db import DB, SafeQuery, User, UserRoles
+from ..db import DB, SafeQuery, User
 from ..knobs import Knob
 from ..metrics import APIMetrics
 from . import errors
@@ -39,7 +42,8 @@ PROVISIONING_AUTH = [
 ]
 
 ADMIN_AUTH = [
-    re.compile(r'(?:/v\d+\.\d+\.\d+)?/users(?:/.+)?')
+    re.compile(r'(?:/v\d+\.\d+\.\d+)?/users(?:/.+)?'),
+    re.compile(r'(?:/v\d+\.\d+\.\d+)?/_status(?:/.+)?')
 ]
 
 
@@ -84,7 +88,7 @@ AUTH_CTX_HEADER = 'x-auth-ctx'
 
 
 def matches_path(request: Request, patterns: List[Pattern[str]]) -> bool:
-    return any((pattern.match(request.path) for pattern in patterns))
+    return any(pattern.match(request.path) for pattern in patterns)
 
 
 @dataclasses.dataclass
@@ -251,7 +255,7 @@ class AuthContext:
             if matches_path(self.request, ADMIN_AUTH):
                 self.verify_auth_basic(session, 'admin')
 
-                if self.user and self.is_authenticated and self.user.role == UserRoles.ADMIN:
+                if self.user and self.is_authenticated and self.user.is_admin:
                     self.is_authorized = True
                     return
 
