@@ -1,17 +1,24 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
+import contextlib
+import dataclasses
 import datetime
 import enum
 import functools
 import hashlib
+import json
 import logging
 import secrets
 import threading
 import time
 from contextlib import contextmanager
-from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, Iterator, List, Optional, Tuple, Type, TypeVar, Union, \
-    cast
+from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, Generic, Iterator, List, Optional, Tuple, Type, \
+    TypeVar, Union, cast
 
 import gluetool.glue
 import gluetool.log
+import psycopg2.errors
 import sqlalchemy
 import sqlalchemy.event
 import sqlalchemy.ext.declarative
@@ -20,17 +27,19 @@ from gluetool.result import Error, Ok, Result
 from sqlalchemy import JSON, Boolean, Column, DateTime, Enum, ForeignKey, Integer, String, Text
 from sqlalchemy.orm import relationship
 from sqlalchemy.orm.query import Query as _Query
+from sqlalchemy.orm.session import sessionmaker
 from sqlalchemy_utils import EncryptedType
 from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine
 
 from .guest import GuestState
 from .knobs import get_vault_password
 
-if TYPE_CHECKING:  # noqa
+if TYPE_CHECKING:
     from mypy_extensions import VarArg
 
     from . import Failure
     from .environment import Environment
+    from .tasks import Actor, ActorArgumentType
 
 
 # Type variables for use in our generic types
@@ -73,13 +82,13 @@ Base = sqlalchemy.ext.declarative.declarative_base()
 # value type.
 #
 # Note: unfortunately, I wasn't able to cover "update" methods - filter(), limit(), ... - with a single
-# type/decorator and preserve the signature enough to keep type checking. I always foud out that methods
+# decorator and preserve the signature enough to keep type checking. I always found out that methods
 # accepting 1 integer parameter (limit/offset) collide with signature of filter() which accepts variable
 # number of arguments: `[argument: int]` does not fit under [*args: Any]`. Therefore we have a decorator
 # for methods accepting anything, and another decorator for methods accepting an integer. On the plus
 # side, type checking works, and mypy does see SafeQuery.limit() as accepting one integer and nothing else.
-SafeQueryRawUpdateVAType = Callable[['SafeQuery[T]', 'VarArg(Any)'], _Query]  # type: ignore
-SafeQueryUpdateVAType = Callable[['SafeQuery[T]', 'VarArg(Any)'], 'SafeQuery[T]']  # type: ignore
+SafeQueryRawUpdateVAType = Callable[['SafeQuery[T]', 'VarArg(Any)'], _Query]  # type: ignore[valid-type]
+SafeQueryUpdateVAType = Callable[['SafeQuery[T]', 'VarArg(Any)'], 'SafeQuery[T]']  # type: ignore[valid-type]
 
 SafeQueryRawUpdateIType = Callable[['SafeQuery[T]', int], _Query]
 SafeQueryUpdateIType = Callable[['SafeQuery[T]', int], 'SafeQuery[T]']
@@ -233,6 +242,101 @@ def stringify_query(session: sqlalchemy.orm.session.Session, query: Any) -> str:
     return str(query.compile(dialect=session.bind.dialect))
 
 
+def execute_db_statement(
+    logger: gluetool.log.ContextAdapter,
+    session: sqlalchemy.orm.session.Session,
+    statement: Union[sqlalchemy.insert, sqlalchemy.update, sqlalchemy.delete],
+    expected_records: Union[int, Tuple[int, int]] = 1,
+) -> Result[Optional[Any], 'Failure']:
+    """
+    Execute a given SQL query, ``INSERT``, ``UPDATE`` or ``DELETE``.
+
+    .. note::
+
+       This routine should replace :py:func:`safe_db_change` one day: we use ``safe_db_change()`` to perform
+       changes that must be synchronized with other actions, like dispatching tasks, which is something
+       we are moving to different implementation based on transaction outbox.
+    """
+
+    from . import Failure
+
+    def to_failure(exc: Exception) -> Failure:
+        if isinstance(exc, sqlalchemy.exc.StatementError):
+            return Failure.from_exc(
+                'failed to execute DB statement',
+                exc,
+                statement=exc.statement,
+                serialization_failure=False
+            )
+
+        return Failure.from_exc(
+            'failed to execute DB statement',
+            exc,
+            serialization_failure=False
+        )
+
+    try:
+        result = cast(
+            sqlalchemy.engine.ResultProxy,
+            session.execute(statement)
+        )
+
+        if result.is_insert:
+            # TODO: INSERT sets this correctly, but what about INSERT + ON CONFLICT? If the row exists,
+            # TODO: rowcount is set to 0, but the (optional) UPDATE did happen, so... UPSERT should probably
+            # TODO: be ready to accept both 0 and 1. We might need to return more than just true/false for
+            # TODO: ON CONFLICT to become auditable.
+            affected_rows = result.rowcount
+
+        else:
+            affected_rows = result.rowcount
+
+        if isinstance(expected_records, tuple) \
+           and not (expected_records[0] <= affected_rows <= expected_records[1]):
+            return Error(Failure(
+                'unexpected number of affected rows',
+                statement=stringify_query(session, statement),
+                serialization_failure=False,
+                affected_rows=affected_rows,
+                expected_affected_rows_min=expected_records[0],
+                expected_affected_rows_max=expected_records[1]
+            ))
+
+        elif affected_rows != expected_records:
+            return Error(Failure(
+                'unexpected number of affected rows',
+                statement=stringify_query(session, statement),
+                serialization_failure=False,
+                affected_rows=affected_rows,
+                expected_affected_rows=expected_records
+            ))
+
+        logger.debug(f'found {affected_rows} matching rows, as expected')
+
+    except sqlalchemy.exc.OperationalError as exc:
+        if not isinstance(exc.orig, psycopg2.errors.SerializationFailure) \
+           or exc.orig.pgerror.strip() != 'ERROR:  could not serialize access due to concurrent update':
+
+            return Error(to_failure(exc))
+
+        return Error(Failure(
+            'failed to execute DB statement',
+            query=exc.statement,
+            serialization_failure=True
+        ))
+
+    except Exception as exc:
+        return Error(to_failure(exc))
+
+    else:
+        if result.is_insert:
+            logger.debug(f'created record with primary key {result.inserted_primary_key}')
+
+            return Ok(result.inserted_primary_key)
+
+        return Ok(None)
+
+
 def safe_db_change(
     logger: gluetool.log.ContextAdapter,
     session: sqlalchemy.orm.session.Session,
@@ -352,7 +456,7 @@ def upsert(
     """
 
     if session.bind.dialect.name != 'postgresql':
-        raise gluetool.glue.GlueError('UPSERT is not support for dialect "{}"'.format(session.bind.dialect.name))
+        raise gluetool.glue.GlueError(f'UPSERT is not support for dialect "{session.bind.dialect.name}"')
 
     from sqlalchemy.dialects.postgresql import insert
 
@@ -389,7 +493,7 @@ def upsert(
 
     if update_data is None:
         statement = statement.on_conflict_do_nothing(
-            constraint=model.__table__.primary_key  # type: ignore
+            constraint=model.__table__.primary_key  # type: ignore[attr-defined]
         )
 
         # INSERT part of the query is still valid, but there's no ON CONFLICT UPDATE... Unfortunatelly,
@@ -398,7 +502,7 @@ def upsert(
 
     else:
         statement = statement.on_conflict_do_update(
-            constraint=model.__table__.primary_key,  # type: ignore
+            constraint=model.__table__.primary_key,  # type: ignore[attr-defined]
             set_=update_data,
             where=where
         )
@@ -408,6 +512,112 @@ def upsert(
     return safe_db_change(logger, session, statement, expected_records=expected_records)
 
 
+@dataclasses.dataclass
+class TransactionResult:
+    success: bool = False
+
+    failure: Optional['Failure'] = None
+    failed_query: Optional[str] = None
+
+
+@contextlib.contextmanager
+def transaction() -> Generator[TransactionResult, None, None]:
+    """
+    Thin context manager for handling possible transation rollback when executing multiple queries.
+
+    .. note::
+
+       Starting a DB session is **not** the responsibility of this context manager - that is left to caller, because
+       such a session can and would be used for queries that do not necessarily need to cause transaction rollback.
+
+    .. code-block:: python
+
+       with DB.get_session(transactional=True) as session:
+           with transaction() as result:
+               session.execute(sqlalchemy.insert(...))
+               session.execute(sqlalchemy.insert(...))
+               session.execute(sqlalchemy.insert(...))
+
+        if result.success is not True:
+            # handle transaction rollback (or DB error)
+            ...
+    """
+
+    result = TransactionResult()
+
+    def _save_error(exc: Exception) -> None:
+        result.success = False
+
+        if isinstance(exc, sqlalchemy.exc.StatementError):
+            result.failure = Failure.from_exc(
+                'failed to execute in transaction',
+                exc=exc,
+                query=exc.statement
+            )
+
+        else:
+            result.failure = Failure.from_exc(
+                'failed to execute in transaction',
+                exc=exc
+            )
+
+    try:
+        yield result
+
+    except sqlalchemy.exc.OperationalError as exc:
+        if not isinstance(exc.orig, psycopg2.errors.SerializationFailure) \
+           or exc.orig.pgerror.strip() != 'ERROR:  could not serialize access due to concurrent update':
+            _save_error(exc)
+            return
+
+        result.success = False
+        result.failed_query = exc.statement
+
+    except Exception as exc:
+        _save_error(exc)
+
+    else:
+        result.success = True
+
+
+def execute_in_transaction(
+    logger: gluetool.log.ContextAdapter,
+    session: sqlalchemy.orm.session.Session,
+    *queries: Union[sqlalchemy.insert, sqlalchemy.update]
+) -> Result[bool, 'Failure']:
+    """
+    Execute given SQL queries as if they were part of a single transaction, and detect transaction rollback.
+
+    .. note::
+
+       Since the DB session is an input parameter of this function, it is **not** the responsibility of this
+       function to start a transaction-aware DB session - that is left to caller, because such a session
+       can and would be used for other queries as well.
+
+    :returns: ``True`` when all queries were successfully executed, ``False`` otherwise.
+    """
+
+    try:
+        for query in queries:
+            session.execute(query)
+
+    except sqlalchemy.exc.OperationalError as exc:
+        if not isinstance(exc.orig, psycopg2.errors.SerializationFailure) \
+           or exc.orig.pgerror.strip() != 'ERROR:  could not serialize access due to concurrent update':
+            raise exc
+
+        return Ok(False)
+
+    except Exception as exc:
+        return Error(Failure.from_exc(
+            'failed to execute in transaction',
+            exc=exc,
+            query=stringify_query(session, query)
+        ))
+
+    return Ok(True)
+
+
 class UserRoles(enum.Enum):
     """
     Possible user roles.
@@ -490,6 +700,10 @@ class User(Base):
             role=role.value
         )
 
+    @property
+    def is_admin(self) -> bool:
+        return self.role == UserRoles.ADMIN  # type: ignore[comparison-overlap]
+
 
 class SSHKey(Base):
     __tablename__ = 'sshkeys'
@@ -530,6 +744,67 @@ class Pool(Base):
     guests = relationship('GuestRequest', back_populates='pool')
 
 
+class TaskRequest(Base):
+    """
+    A request to run a task by dispatcher.
+
+    Dispatching a task immediately, e.g. with :py:func:`tft.artemis.tasks.dispatch_task`, might be easier
+    but also prone to condition known as "dual write": when task is a follow-up of a DB change, it's impossible
+    to guarantee that either both (the DB change and message dispatch) or none of those are performed, with DB
+    or broker transactions only.
+
+    A solution most suitable to us is called "transaction outbox": instead of dispatching a task, a DB record
+    is created within the same transaction as the original DB change. These "task requests" are then read by
+    another service that transforms them into actual broker messages, removing the DB record after successfull
+    dispatch.
+
+    Thanks to this split, we can guarantee a both DB change and the need for a follow-up message are safely stored.
+
+    * https://microservices.io/patterns/data/transactional-outbox.html
+    """
+
+    __tablename__ = 'task_requests'
+
+    id = Column(Integer(), primary_key=True, autoincrement=True)
+    taskname = Column(Text(), nullable=False)
+    arguments = Column(JSON(), nullable=False)
+    delay = Column(Integer(), nullable=True)
+
+    @classmethod
+    def create_query(
+        cls,
+        task: 'Actor',
+        *args: 'ActorArgumentType',
+        delay: Optional[int] = None
+    ) -> sqlalchemy.insert:
+        """
+        """
+
+        return sqlalchemy.insert(cls.__table__).values(
+            taskname=task.actor_name,
+            arguments=list(args),
+            delay=delay
+        )
+
+    @classmethod
+    def create(
+        cls,
+        logger: gluetool.log.ContextAdapter,
+        session: sqlalchemy.orm.session.Session,
+        task: 'Actor',
+        *args: 'ActorArgumentType',
+        delay: Optional[int] = None
+    ) -> Result[int, 'Failure']:
+        stmt = cls.create_query(task, *args, delay=delay)
+
+        r = execute_db_statement(logger, session, stmt)
+
+        if r.is_error:
+            return Error(r.unwrap_error())
+
+        return Ok(cast(Tuple[int], r.unwrap())[0])
+
+
 UserDataType = Dict[str, Optional[str]]
 
 
@@ -547,7 +822,7 @@ class GuestRequest(Base):
         # avoid circular imports
         from .environment import Environment
 
-        return Environment.unserialize_from_json(cast(Dict[str, Any], self._environment))
+        return Environment.unserialize(cast(Dict[str, Any], self._environment))
 
     # This is tricky:
     # * we want to keep `nullable=False`, because `ctime` cannot ever be set to `NULL`. That way we're not forced
@@ -577,6 +852,7 @@ class GuestRequest(Base):
     ctime = Column(DateTime(), nullable=False, default=datetime.datetime.utcnow)
 
     state = Column(Enum(GuestState), nullable=False)
+    state_mtime = Column(DateTime(), nullable=True)
 
     address = Column(String(250), nullable=True)
 
@@ -614,6 +890,45 @@ class GuestRequest(Base):
     priority_group = relationship('PriorityGroup', back_populates='guests')
     pool = relationship('Pool', back_populates='guests')
 
+    @classmethod
+    def create_query(
+        cls,
+        guestname: str,
+        environment: 'Environment',
+        ownername: str,
+        ssh_keyname: str,
+        ssh_port: int,
+        ssh_username: str,
+        priorityname: Optional[str],
+        user_data: Optional[UserDataType],
+        skip_prepare_verify_ssh: bool,
+        post_install_script: Optional[str],
+        log_types: List[Tuple[str, GuestLogContentType]]
+    ) -> sqlalchemy.insert:
+        return sqlalchemy.insert(cls.__table__).values(
+            guestname=guestname,
+            _environment=environment.serialize(),
+            ownername=ownername,
+            ssh_keyname=ssh_keyname,
+            ssh_port=ssh_port,
+            ssh_username=ssh_username,
+            priorityname=priorityname,
+            _user_data=user_data,
+            skip_prepare_verify_ssh=skip_prepare_verify_ssh,
+            post_install_script=post_install_script,
+            _log_types=[
+                {
+                    'logtype': log[0],
+                    'contenttype': log[1].value
+                }
+                for log in log_types
+            ],
+            state=GuestState.ROUTING,
+            state_mtime=datetime.datetime.utcnow(),
+            poolname=None,
+            pool_data=json.dumps({}),
+        )
+
     @classmethod
     def log_event_by_guestname(
         cls,
@@ -636,7 +951,7 @@ class GuestRequest(Base):
         r = safe_db_change(
             logger,
             session,
-            sqlalchemy.insert(GuestEvent.__table__).values(  # type: ignore  # GuestEvent *has* __table__
+            sqlalchemy.insert(GuestEvent.__table__).values(  # type: ignore[attr-defined]
                 guestname=guestname,
                 eventname=eventname,
                 _details=details
@@ -812,10 +1127,6 @@ class GuestRequest(Base):
             **details
         )
 
-    @property
-    def is_promised(self) -> bool:
-        return self.address is None
-
     def fetch_events(
         self,
         session: sqlalchemy.orm.session.Session,
@@ -883,8 +1194,8 @@ class GuestEvent(Base):
     guestname = Column(String(250), nullable=False, index=True)
     eventname = Column(String(250), nullable=False)
 
-    # Details are stored as JSON blob, in a "hidden" column - when accessing event details, we'd like them to be
-    # typed correctly, and there will never ever be an event having a list or an integer as a detail, it will always
+    # Details are stored as JSON blob, in a "hidden" column - when accessing event details, we'd like to cast them to
+    # proper type, and there will never ever be an event having a list or an integer as a detail, it will always
     # be a mapping. Therefore `_details` column and `details` property to apply proper cast call.
     _details = Column(JSON(), nullable=False, server_default='{}')
 
@@ -1022,7 +1333,7 @@ class Knob(Base):
 # TODO: shuffle a bit with files to avoid local imports and to set this up conditionaly. It's probably not
 # critical, but it would also help a bit with DB class, and it would be really nice to not install the event
 # hooks when not asked to log slow queries.
-@sqlalchemy.event.listens_for(sqlalchemy.engine.Engine, 'before_cursor_execute')  # type: ignore  # untyped decorator
+@sqlalchemy.event.listens_for(sqlalchemy.engine.Engine, 'before_cursor_execute')  # type: ignore[no-untyped-call,misc]
 def before_cursor_execute(
     conn: sqlalchemy.engine.Connection,
     cursor: Any,
@@ -1034,7 +1345,7 @@ def before_cursor_execute(
     conn.info.setdefault('query_start_time', []).append(time.time())
 
 
-@sqlalchemy.event.listens_for(sqlalchemy.engine.Engine, 'after_cursor_execute')  # type: ignore  # untyped decorator
+@sqlalchemy.event.listens_for(sqlalchemy.engine.Engine, 'after_cursor_execute')  # type: ignore[no-untyped-call,misc]
 def after_cursor_execute(
     conn: sqlalchemy.engine.Connection,
     cursor: Any,
@@ -1067,6 +1378,21 @@ class DB:
     instance: Optional['DB'] = None
     _lock = threading.RLock()
 
+    #: "Root" engine, with the setup dictated by DB configuration.
+    engine: sqlalchemy.engine.Engine
+
+    #: Engine derived from :py:attr:`engine`, configured to use no transactions in an auto-commit fashion.
+    engine_autocommit: sqlalchemy.engine.Engine
+    #: Session factory on top of auto-commit :py:attr:`engine_autocommit` engine: ``AUTOCOMMIT`` isolation level
+    #: is applied to each and every query.
+    sessionmaker_autocommit: sessionmaker
+
+    #: Engine derived from :py:attr:`engine`, configured to transactions with ``REPEATABLE READ`` isolation level.
+    engine_transactional: sqlalchemy.engine.Engine
+    #: Session factory on top of auto-commit :py:attr:`engine_treansactional` engine: every session spawns
+    #: a transaction with ``REPEATABLE READ`` isolation level.
+    sessionmaker_transactional: sessionmaker
+
     def _setup_instance(
         self,
         logger: gluetool.log.ContextAdapter,
@@ -1077,7 +1403,7 @@ class DB:
 
         self.logger = logger
 
-        logger.info('connecting to db {}'.format(url))
+        logger.info(f'connecting to db {url}')
 
         if KNOB_LOGGING_DB_QUERIES.value:
             gluetool.log.Logging.configure_logger(logging.getLogger('sqlalchemy.engine'))
@@ -1116,7 +1442,17 @@ class DB:
         else:
             self.engine = sqlalchemy.create_engine(url)
 
-        self._sessionmaker = sqlalchemy.orm.sessionmaker(bind=self.engine)
+        # TODO: hopefully, with better sqlalchemy stubs, cast() wouldn't be needed anymore
+        _engine_execution_options = cast(
+            Callable[..., sqlalchemy.engine.Engine],
+            self.engine.execution_options
+        )
+
+        self.engine_autocommit = _engine_execution_options(isolation_level='AUTOCOMMIT')
+        self.sessionmaker_autocommit = sqlalchemy.orm.sessionmaker(bind=self.engine_autocommit)
+
+        self.engine_transactional = _engine_execution_options(isolation_level='REPEATABLE READ')
+        self.sessionmaker_transactional = sqlalchemy.orm.sessionmaker(bind=self.engine_transactional)
 
     def __new__(
         cls,
@@ -1126,14 +1462,26 @@ class DB:
     ) -> 'DB':
         with cls._lock:
             if cls.instance is None:
-                cls.instance = super(DB, cls).__new__(cls)
+                cls.instance = super().__new__(cls)
                 cls.instance._setup_instance(logger, url, application_name=application_name)
 
         return cls.instance
 
     @contextmanager
-    def get_session(self) -> Iterator[sqlalchemy.orm.session.Session]:
-        Session = sqlalchemy.orm.scoped_session(self._sessionmaker)
+    def get_session(self, transactional: bool = False) -> Iterator[sqlalchemy.orm.session.Session]:
+        """
+        Create new DB session.
+
+        :param transactional: if set, session will support transactions rather than auto-commit isolation level.
+        :returns: new DB session.
+        """
+
+        if transactional:
+            Session = sqlalchemy.orm.scoped_session(self.sessionmaker_transactional)
+
+        else:
+            Session = sqlalchemy.orm.scoped_session(self.sessionmaker_autocommit)
+
         session = Session()
 
         if self._echo_pool:
@@ -1142,7 +1490,7 @@ class DB:
             gluetool.log.log_dict(
                 self.logger.info,
                 'pool metrics',
-                DBPoolMetrics.load(self.logger, self, session)  # type: ignore
+                DBPoolMetrics.load(self.logger, self, session)  # type: ignore[attr-defined]
             )
 
         try:
