@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import contextlib
 import dataclasses
 import datetime
@@ -10,7 +13,8 @@ import sys
 import tempfile
 import threading
 import time
-from typing import Any, Callable, Dict, Generic, Iterator, List, Optional, Pattern, Tuple, Type, TypeVar, Union, cast
+from typing import Any, Callable, Dict, Generic, Iterable, Iterator, List, Optional, Pattern, Tuple, Type, TypeVar, \
+    Union, cast
 
 import gluetool
 import gluetool.log
@@ -18,14 +22,16 @@ import gluetool.utils
 import sqlalchemy
 import sqlalchemy.orm.session
 from gluetool.result import Error, Ok, Result
-from typing_extensions import Protocol, TypedDict
+from pint import Quantity
+from typing_extensions import Literal, Protocol, TypedDict
 
-from .. import Failure, JSONType, SerializableContainer, get_cached_item, get_cached_items_as_list, log_dict_yaml, \
-    process_output_to_str, refresh_cached_set, safe_call
+from .. import Failure, JSONType, SerializableContainer, log_dict_yaml, process_output_to_str, safe_call
+from ..cache import get_cached_set_as_list, get_cached_set_item, refresh_cached_set
 from ..context import CACHE, LOGGER
-from ..db import GuestLog, GuestLogContentType, GuestLogState, GuestRequest, GuestTag, SnapshotRequest, SSHKey
-from ..environment import UNITS, Environment, Flavor, FlavorDisk, FlavorDisks, MeasurableConstraintValueType
-from ..knobs import Knob
+from ..db import GuestLog, GuestLogContentType, GuestLogState, GuestRequest, GuestTag, Pool, SafeQuery, \
+    SnapshotRequest, SSHKey
+from ..environment import UNITS, Environment, Flavor, FlavorBoot, FlavorDisk, FlavorDisks, MeasurableConstraintValueType
+from ..knobs import KNOB_POOL_ENABLED, Knob
 from ..metrics import PoolCostsMetrics, PoolMetrics, PoolResourcesMetrics, ResourceType
 from ..script import hook_engine
 
@@ -51,11 +57,30 @@ ConfigFlavorCPUSpecType = TypedDict(
 )
 
 
-#: pools[].parameters.{custom-flavors,patch-flavors}[].disk
-class ConfigFlavorDiskSpecType(TypedDict):
+#: pools[].parameters.{custom-flavors,patch-flavors}[].disk (static disk)
+class ConfigFlavorDiskSpecificSpecType(TypedDict):
     size: Optional[Union[int, str]]
 
 
+#: pools[].parameters.{custom-flavors,patch-flavors}[].disk (expansion)
+ConfigFlavorDiskExpansionSpecType = TypedDict(
+    'ConfigFlavorDiskExpansionSpecType',
+    {
+        'max-count': int,
+        'min-size': Optional[Union[int, str]],
+        'max-size': Optional[Union[int, str]],
+    }
+)
+
+ConfigFlavorDiskExpandedSpecType = TypedDict(
+    'ConfigFlavorDiskExpandedSpecType',
+    {
+        'additional-disks': ConfigFlavorDiskExpansionSpecType
+    }
+)
+
+ConfigFlavorDiskSpecType = Union[ConfigFlavorDiskSpecificSpecType, ConfigFlavorDiskExpandedSpecType]
+
 #: pools[].parameters.{custom-flavors,patch-flavors}[].virtualization
 ConfigFlavorVirtualizationSpecType = TypedDict(
     'ConfigFlavorVirtualizationSpecType',
@@ -73,6 +98,7 @@ ConfigPatchFlavorSpecType = TypedDict(
     {
         'name': str,
         'name-regex': str,
+        'arch': str,
         'cpu': ConfigFlavorCPUSpecType,
         'disk': List[ConfigFlavorDiskSpecType],
         'virtualization': ConfigFlavorVirtualizationSpecType
@@ -81,16 +107,14 @@ ConfigPatchFlavorSpecType = TypedDict(
 
 
 #: pools[].parameters.custom-flavors[]
-ConfigCustomFlavorSpecType = TypedDict(
-    'ConfigCustomFlavorSpecType',
-    {
-        'name': str,
-        'base': str,
-        'cpu': ConfigFlavorCPUSpecType,
-        'disk': List[ConfigFlavorDiskSpecType],
-        'virtualization': ConfigFlavorVirtualizationSpecType
-    }
-)
+class ConfigCustomFlavorSpecType(TypedDict):
+    name: str
+    base: str
+    arch: str
+    cpu: ConfigFlavorCPUSpecType
+    disk: List[ConfigFlavorDiskSpecType]
+    virtualization: ConfigFlavorVirtualizationSpecType
+
 
 ConfigFlavorSpecType = Union[ConfigPatchFlavorSpecType, ConfigCustomFlavorSpecType]
 
@@ -110,8 +134,10 @@ ConfigImageSpecType = TypedDict(
     }
 )
 
-ConfigCapabilitiesDisableGuestLog = TypedDict(
-    'ConfigCapabilitiesDisableGuestLog',
+
+#: pools[].capabilities.disable-guest-logs
+ConfigCapabilitiesDisableGuestLogType = TypedDict(
+    'ConfigCapabilitiesDisableGuestLogType',
     {
         'log-name': str,
         'content-type': str
@@ -119,6 +145,18 @@ ConfigCapabilitiesDisableGuestLog = TypedDict(
 )
 
 
+#: pools[].capabilities
+ConfigCapabilitiesType = TypedDict(
+    'ConfigCapabilitiesType',
+    {
+        'supported-architectures': Union[Literal['any'], List[str]],
+        'supports-snapshots': Union[str, bool],
+        'supports-spot-instances': Union[str, bool],
+        'disable-guest-logs': List[ConfigCapabilitiesDisableGuestLogType]
+    }
+)
+
+
 IP_ADDRESS_PATTERN = re.compile(r'((?:[0-9]{1,3}\.){3}[0-9]{1,3})')  # noqa: FS003
 
 
@@ -226,11 +264,16 @@ except Exception as exc:
 
 
 if hasattr(shlex, 'join'):
-    command_join = shlex.join  # type: ignore
+    # We cannot use `type: ignore` comment here, because it applies to Python 3.7 only, where `shlex.join`
+    # does not exist. In newer Python versions, is does and therefore the ignore hint is reported as unused.
+    # And there is no way to disable *that* report :/ See https://github.com/python/mypy/issues/8823
+    # Trying to work around this with a bit of `getattr()` - we *know* the attribute exists, and the types
+    # are a match, we just need to fool mypy a bit.
+    command_join = getattr(shlex, 'join')
 
 else:
-    def command_join(command: List[str]) -> str:
-        return ' '.join(shlex.quote(arg) for arg in command)
+    def command_join(split_command: Iterable[str]) -> str:
+        return ' '.join(shlex.quote(arg) for arg in split_command)
 
 
 class _AnyArchitecture:
@@ -257,19 +300,35 @@ class CLIOutput:
 
 class PoolLogger(gluetool.log.ContextAdapter):
     def __init__(self, logger: gluetool.log.ContextAdapter, poolname: str) -> None:
-        super(PoolLogger, self).__init__(logger, {
+        super().__init__(logger, {
             'ctx_pool_name': (10, poolname)
         })
 
+    @property
+    def poolname(self) -> str:
+        return cast(str, self._contexts['pool_name'][1])
+
+
+class CLIErrorCauses(enum.Enum):
+    """
+    A base class for enums listing various error causes recognized by pools for their CLI tools.
+    """
+
+    pass
+
+
+#: A type for callables extracting CLI error cause from process output.
+CLIErrorCauseExtractor = Callable[
+    [gluetool.utils.ProcessOutput],
+    CLIErrorCauses
+]
+
 
 @dataclasses.dataclass(repr=False)
 class PoolImageSSHInfo(SerializableContainer):
     username: str = 'root'
     port: int = 22
 
-    def __repr__(self) -> str:
-        return f'<PoolImageSSHInfo: username={self.username} port={self.port}>'
-
 
 @dataclasses.dataclass(repr=False)
 class PoolImageInfo(SerializableContainer):
@@ -292,38 +351,10 @@ class PoolImageInfo(SerializableContainer):
     name: str
     id: str
 
+    boot: FlavorBoot
     ssh: PoolImageSSHInfo
 
-    def __repr__(self) -> str:
-        return f'<PoolImageInfo: name={self.name} id={self.id} ssh={repr(self.ssh)}>'
-
-    def serialize_to_json(self) -> Dict[str, Any]:
-        """
-        Serialize properties to JSON.
-
-        :returns: serialized form of image properties.
-        """
-
-        serialized = super(PoolImageInfo, self).serialize_to_json()
-        serialized['ssh'] = self.ssh.serialize_to_json()
-
-        return serialized
-
-    @classmethod
-    def unserialize_from_json(cls, serialized: Dict[str, Any]) -> 'PoolImageInfo':
-        """
-        Unserialize properties from JSON.
-
-        :param serialized: serialized form of image properties.
-        :returns: image info instance.
-        """
-
-        image = cls(**serialized)
-        image.ssh = PoolImageSSHInfo.unserialize_from_json(serialized['ssh'])
-
-        return image
-
-    def serialize_to_json_scrubbed(self) -> Dict[str, Any]:
+    def serialize_scrubbed(self) -> Dict[str, Any]:
         """
         Serialize properties to JSON while scrubbing sensitive information.
 
@@ -378,6 +409,9 @@ class PoolCapabilities:
     #: List of log name/log content type pairs describing that logs are supported by the driver.
     supported_guest_logs: List[Tuple[str, GuestLogContentType]] = dataclasses.field(default_factory=list)
 
+    #: If set, the pool can find instances by their hostnames.
+    supports_hostnames: bool = False
+
     def supports_arch(self, arch: str) -> bool:
         """
         Check whether a given architecture is supported. It is either listed among architectures supported
@@ -423,7 +457,7 @@ class PoolData:
 
     @classmethod
     def unserialize(cls: Type[T], guest_request: GuestRequest) -> T:
-        return cls(**json.loads(guest_request.pool_data))  # type: ignore
+        return cls(**json.loads(guest_request.pool_data))
 
 
 class ProvisioningState(enum.Enum):
@@ -493,8 +527,8 @@ class PoolResourcesIDs(SerializableContainer):
     def is_empty(self) -> bool:
         return all([value is None for key, value in dataclasses.asdict(self).items() if key != 'ctime'])
 
-    def serialize_to_json(self) -> Dict[str, Any]:
-        serialized = super(PoolResourcesIDs, self).serialize_to_json()
+    def serialize(self) -> Dict[str, Any]:
+        serialized = super().serialize()
 
         # Convert datetime object to string
         serialized['ctime'] = serialized['ctime'].strftime(RESOURCE_CTIME_FMT) if serialized['ctime'] else None
@@ -502,20 +536,54 @@ class PoolResourcesIDs(SerializableContainer):
         return serialized
 
     @classmethod
-    def unserialize_from_json(cls: Type[S], serialized: Dict[str, Any]) -> S:
+    def unserialize(cls: Type[S], serialized: Dict[str, Any]) -> S:
         # Convert ctime string to datetime object
         serialized['ctime'] = datetime.datetime.strptime(serialized['ctime'], RESOURCE_CTIME_FMT) \
             if serialized['ctime'] else None
 
         return cls(**serialized)
 
-    # Adding (un)serialize methods because we have a dedicated serialized type.
-    def serialize(self) -> SerializedPoolResourcesIDs:
-        return self.serialize_to_str()
+    # Overwriting (un)serialize methods because we have a dedicated serialized type.
+    def serialize_to_json(self) -> SerializedPoolResourcesIDs:
+        return super().serialize_to_json()
 
     @classmethod
-    def unserialize(cls: Type[S], serialized: SerializedPoolResourcesIDs) -> S:
-        return cls.unserialize_from_str(serialized)
+    def unserialize_from_json(cls: Type[S], serialized: SerializedPoolResourcesIDs) -> S:
+        return super().unserialize_from_json(serialized)
+
+
+def _parse_flavor_disk_size(
+    field_name: str,
+    value: Optional[Union[str, int]],
+    disk: FlavorDisk
+) -> Result[Optional[Quantity], Failure]:
+    if value is None:
+        return Ok(None)
+
+    property_name = field_name.replace('-', '_')
+
+    r_value = safe_call(UNITS, value)
+
+    if r_value.is_error:
+        return Error(Failure.from_failure(
+            f'failed to parse flavor disk.{field_name}',
+            r_value.unwrap_error(),
+            details={
+                property_name: value
+            }
+        ))
+
+    raw_value = r_value.unwrap()
+
+    if isinstance(raw_value, int):
+        real_value = UNITS.Quantity(raw_value, UNITS.bytes)
+
+    else:
+        real_value = raw_value
+
+    setattr(disk, property_name, real_value)
+
+    return Ok(real_value)
 
 
 def _apply_flavor_specification(
@@ -529,6 +597,9 @@ def _apply_flavor_specification(
     fields.
     """
 
+    if 'arch' in flavor_spec:
+        flavor.arch = flavor_spec['arch']
+
     if 'cpu' in flavor_spec:
         cpu_patch = flavor_spec['cpu']
 
@@ -553,23 +624,41 @@ def _apply_flavor_specification(
             disk = FlavorDisk()
             patched_disks.append(disk)
 
-            if 'size' in disk_patch:
-                r_space = safe_call(UNITS, disk_patch['size'])
+            if 'additional-disks' not in disk_patch:
+                specific_disk_patch = cast(ConfigFlavorDiskSpecificSpecType, disk_patch)
 
-                if r_space.is_error:
-                    return Error(Failure.from_failure(
-                        'failed to parse flavor disk.size',
-                        r_space.unwrap_error(),
-                        space=disk_patch['size']
-                    ))
+                r_size = _parse_flavor_disk_size(
+                    'size',
+                    specific_disk_patch.get('size'),
+                    disk
+                )
 
-                raw_size = r_space.unwrap()
+                if r_size.is_error:
+                    return Error(r_size.unwrap_error())
 
-                if isinstance(raw_size, int):
-                    disk.size = raw_size * UNITS['bytes']
+            else:
+                expansion_disk_patch = cast(ConfigFlavorDiskExpandedSpecType, disk_patch)['additional-disks']
 
-                else:
-                    disk.size = raw_size
+                disk.is_expansion = True
+                disk.max_additional_items = expansion_disk_patch['max-count']
+
+                r_size = _parse_flavor_disk_size(
+                    'min-size',
+                    expansion_disk_patch['min-size'],
+                    disk
+                )
+
+                if r_size.is_error:
+                    return Error(r_size.unwrap_error())
+
+                r_size = _parse_flavor_disk_size(
+                    'max-size',
+                    expansion_disk_patch['max-size'],
+                    disk
+                )
+
+                if r_size.is_error:
+                    return Error(r_size.unwrap_error())
 
         flavor.disk = FlavorDisks(patched_disks)
 
@@ -809,7 +898,7 @@ class HookImageInfoMapper(ImageInfoMapper[_PoolImageInfoTypeVar]):
     """
 
     def __init__(self, pool: 'PoolDriver', hook_name: str) -> None:
-        super(HookImageInfoMapper, self).__init__(pool)
+        super().__init__(pool)
 
         self.hook_name = hook_name
 
@@ -841,9 +930,38 @@ class HookImageInfoMapper(ImageInfoMapper[_PoolImageInfoTypeVar]):
         return r_image
 
 
+class GuestLogUpdaterType(Protocol):
+    __name__: str
+
+    def __call__(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
+        guest_log: GuestLog
+    ) -> Result[GuestLogUpdateProgress, Failure]:
+        pass
+
+
+def guest_log_updater(
+    drivername: str,
+    logname: str,
+    contenttype: GuestLogContentType
+) -> Callable[[GuestLogUpdaterType], GuestLogUpdaterType]:
+    def wrapper(func: GuestLogUpdaterType) -> GuestLogUpdaterType:
+        PoolDriver.guest_log_updaters[(drivername, logname, contenttype)] = func.__name__
+
+        return func
+
+    return wrapper
+
+
 class PoolDriver(gluetool.log.LoggerMixin):
+    drivername: str
+
     image_info_class: Type[PoolImageInfo] = PoolImageInfo
     flavor_info_class: Type[Flavor] = Flavor
+    pool_data_class: Type[PoolData] = PoolData
+    cli_error_cause_extractor: Optional[CLIErrorCauseExtractor] = None
 
     #: Template for a cache key holding pool image info.
     POOL_IMAGE_INFO_CACHE_KEY = 'pool.{}.image-info'
@@ -851,13 +969,16 @@ class PoolDriver(gluetool.log.LoggerMixin):
     #: Template for a cache key holding flavor image info.
     POOL_FLAVOR_INFO_CACHE_KEY = 'pool.{}.flavor-info'
 
+    #: Hold all known guest log updaters, with key being driver name, log name and its content type.
+    guest_log_updaters: Dict[Tuple[str, str, GuestLogContentType], str] = {}
+
     def __init__(
         self,
         logger: gluetool.log.ContextAdapter,
         poolname: str,
         pool_config: Dict[str, Any]
     ) -> None:
-        super(PoolDriver, self).__init__(logger)
+        super().__init__(logger)
 
         self.poolname = poolname
         self.pool_config = pool_config
@@ -868,6 +989,134 @@ class PoolDriver(gluetool.log.LoggerMixin):
         self.image_info_cache_key = self.POOL_IMAGE_INFO_CACHE_KEY.format(self.poolname)  # noqa: FS002
         self.flavor_info_cache_key = self.POOL_FLAVOR_INFO_CACHE_KEY.format(self.poolname)  # noqa: FS002
 
+    _drivers_registry: Dict[str, Type['PoolDriver']] = {}
+
+    @staticmethod
+    def _instantiate(
+        logger: gluetool.log.ContextAdapter,
+        driver_name: str,
+        poolname: str,
+        pool_config: Dict[str, Any]
+    ) -> Result['PoolDriver', Failure]:
+        pool_driver_class = PoolDriver._drivers_registry.get(driver_name)
+
+        if pool_driver_class is None:
+            return Error(Failure('cannot find pool driver', drivername=driver_name))
+
+        pool = pool_driver_class(logger, poolname, pool_config)
+
+        r_sanity = pool.sanity()
+
+        if r_sanity.is_error:
+            return Error(r_sanity.unwrap_error())
+
+        return Ok(pool)
+
+    # Because sometimes we just don't know whether the given pool actually exists or not...
+    @staticmethod
+    def load_or_none(
+        logger: gluetool.log.ContextAdapter,
+        session: sqlalchemy.orm.session.Session,
+        poolname: str
+    ) -> Result[Optional['PoolDriver'], Failure]:
+        r_pool_record = SafeQuery.from_session(session, Pool) \
+            .filter(Pool.poolname == poolname) \
+            .one_or_none()
+
+        if r_pool_record.is_error:
+            return Error(r_pool_record.unwrap_error())
+
+        pool_record = r_pool_record.unwrap()
+
+        if pool_record is None:
+            return Ok(None)
+
+        r_pool = PoolDriver._instantiate(logger, pool_record.driver, poolname, pool_record.parameters)
+
+        if r_pool.is_error:
+            return Error(r_pool.unwrap_error())
+
+        return Ok(r_pool.unwrap())
+
+        # And when .map()/.map_error() become available, the code above would become much less spaghetti-ish...
+        #
+        # TODO: switch to when map/map_error become available
+        #
+        # def instantiate(pool_record: Optional[Pool]) -> Result[Optional['PoolDriver'], Failure]:
+        #     if pool_record is None:
+        #         return Ok(None)
+        #
+        #     # `_instantiate()`` return either a pool or failure, never `None`. That is nice, but it isn't matching
+        #     # the expected return value of `load_or_none()`, therefore we need a cast.
+        #     return cast(
+        #         Result[Optional['PoolDriver'], Failure],
+        #         PoolDriver._instantiate(pool_record.driver, logger, poolname, pool_record.parameters)
+        #     )
+        #
+        # return SafeQuery.from_session(session, Pool) \
+        #     .filter(Pool.poolname == poolname) \
+        #     .one_or_none() \
+        #     .map(instantiate)
+
+    # ... and sometimes, we are pretty sure the pool does exist, and it's a hard error if we can't find it.
+    @staticmethod
+    def load(
+        logger: gluetool.log.ContextAdapter,
+        session: sqlalchemy.orm.session.Session,
+        poolname: str
+    ) -> Result['PoolDriver', Failure]:
+        r_pool = PoolDriver.load_or_none(logger, session, poolname)
+
+        if r_pool.is_error:
+            return Error(r_pool.unwrap_error())
+
+        pool = r_pool.unwrap()
+
+        if pool is None:
+            return Error(Failure(
+                'no such pool',
+                poolname=poolname
+            ))
+
+        return Ok(pool)
+
+        # TODO: switch to when map/map_error become available
+        #
+        # return PoolDriver.load_or_none(logger, session, poolname) \
+        #     .map(lambda pool: Error(Failure('no such pool', poolname=poolname)) if pool is None else Ok(pool))
+
+    @staticmethod
+    def load_all(
+        logger: gluetool.log.ContextAdapter,
+        session: sqlalchemy.orm.session.Session,
+        enabled_only: bool = True
+    ) -> Result[List['PoolDriver'], Failure]:
+        r_pools = SafeQuery.from_session(session, Pool).all()
+
+        if r_pools.is_error:
+            return Error(r_pools.unwrap_error())
+
+        pools: List[PoolDriver] = []
+
+        for pool_record in r_pools.unwrap():
+            if enabled_only is True:
+                r_enabled = KNOB_POOL_ENABLED.get_value(session=session, poolname=pool_record.poolname)
+
+                if r_enabled.is_error:
+                    return Error(r_enabled.unwrap_error())
+
+                if r_enabled.unwrap() is not True:
+                    continue
+
+            r_pool = PoolDriver._instantiate(logger, pool_record.driver, pool_record.poolname, pool_record.parameters)
+
+            if r_pool.is_error:
+                return Error(r_pool.unwrap_error())
+
+            pools.append(r_pool.unwrap())
+
+        return Ok(pools)
+
     def __repr__(self) -> str:
         return f'<{self.__class__.__name__}: {self.poolname}>'
 
@@ -879,6 +1128,10 @@ class PoolDriver(gluetool.log.LoggerMixin):
 
         raise NotImplementedError()
 
+    @property
+    def use_only_when_addressed(self) -> bool:
+        return self.pool_config.get('use-only-when-addressed', False)
+
     def sanity(self) -> Result[bool, Failure]:
         """
         Do sanity checks after initializing the driver. Useful to check for pool configuration
@@ -915,7 +1168,7 @@ class PoolDriver(gluetool.log.LoggerMixin):
             logger,
             release_pool_resources,
             self.poolname,
-            resource_ids.serialize(),
+            resource_ids.serialize_to_json(),
             guest_request.guestname if guest_request else None,
             delay=r_delay.unwrap()
         )
@@ -969,6 +1222,31 @@ class PoolDriver(gluetool.log.LoggerMixin):
         if not capabilities.supports_arch(guest_request.environment.hw.arch):
             return Ok(False)
 
+        # Check whether given HW constraints do not go against what pool can deliver.
+        #
+        # There may be more checks implemented by the driver, but some conflicts we
+        # can identify right away.
+        if guest_request.environment.has_hw_constraints:
+            r_constraints = guest_request.environment.get_hw_constraints()
+
+            if r_constraints.is_error:
+                return Error(r_constraints.unwrap_error())
+
+            constraints = r_constraints.unwrap()
+
+            assert constraints is not None
+
+            # If request may depend on particular machine hostname to be available
+            # (or not available), pool must support hostnames first.
+            if not capabilities.supports_hostnames:
+                r_uses_hostname = constraints.uses_constraint(logger, 'hostname')
+
+                if r_uses_hostname.is_error:
+                    return Error(r_uses_hostname.unwrap_error())
+
+                if r_uses_hostname.unwrap() is True:
+                    return Ok(False)
+
         return Ok(True)
 
     def _map_image_name_to_image_info_by_cache(
@@ -1006,6 +1284,30 @@ class PoolDriver(gluetool.log.LoggerMixin):
 
         raise NotImplementedError()
 
+    # TODO: I dislike the naming scheme here very much...
+    def _map_environment_to_flavor_info_by_cache_by_name_or_none(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        flavorname: str
+    ) -> Result[Optional[Flavor], Failure]:
+        """
+        Find a flavor matching the given name.
+
+        :returns: a flavor info or ``None`` if such a name does not exist.
+        """
+
+        r_flavor = get_cached_set_item(
+            CACHE.get(),
+            self.flavor_info_cache_key,
+            flavorname,
+            self.flavor_info_class
+        )
+
+        if r_flavor.is_error:
+            return Error(r_flavor.unwrap_error())
+
+        return r_flavor
+
     def _map_environment_to_flavor_info_by_cache_by_name(
         self,
         logger: gluetool.log.ContextAdapter,
@@ -1017,12 +1319,7 @@ class PoolDriver(gluetool.log.LoggerMixin):
         :returns: a flavor info.
         """
 
-        r_flavor = get_cached_item(
-            CACHE.get(),
-            self.flavor_info_cache_key,
-            flavorname,
-            self.flavor_info_class
-        )
+        r_flavor = self._map_environment_to_flavor_info_by_cache_by_name_or_none(logger, flavorname)
 
         if r_flavor.is_error:
             return Error(r_flavor.unwrap_error())
@@ -1056,30 +1353,7 @@ class PoolDriver(gluetool.log.LoggerMixin):
         if r_flavors.is_error:
             return Error(r_flavors.unwrap_error())
 
-        r_capabilities = self.capabilities()
-
-        if r_capabilities.is_error:
-            return Error(r_capabilities.unwrap_error())
-
-        pool_flavors = r_flavors.unwrap()
-        capabilities = r_capabilities.unwrap()
-
-        # For each flavor info and arch, create a Flavor description we can then match against the HW constraints.
-        # TODO: what if some flavors are not supported on all arches?
-        flavors: List[Flavor] = []
-
-        for flavor in pool_flavors:
-            if capabilities.supported_architectures is AnyArchitecture:
-                flavors.append(flavor)
-
-            else:
-                assert isinstance(capabilities.supported_architectures, list)
-
-                for arch in capabilities.supported_architectures:
-                    arch_flavor = flavor.clone()
-                    arch_flavor.arch = arch
-
-                    flavors.append(arch_flavor)
+        flavors = r_flavors.unwrap()
 
         gluetool.log.log_dict(logger.debug, 'available flavors', flavors)
 
@@ -1094,14 +1368,19 @@ class PoolDriver(gluetool.log.LoggerMixin):
         if constraints is None:
             return Ok(flavors)
 
-        gluetool.log.log_blob(logger.debug, 'constraint', constraints.format())  # noqa: FS002
+        log_dict_yaml(logger.debug, 'constraint', constraints.serialize())
 
         # The actual filter: pick flavors that pass the test and match the requirements.
-        suitable_flavors = [
-            flavor
-            for flavor in flavors
-            if constraints.eval_flavor(logger, flavor) is True
-        ]
+        suitable_flavors = []
+
+        for flavor in flavors:
+            r_suitable = constraints.eval_flavor(logger, flavor)
+
+            if r_suitable.is_error:
+                return Error(r_suitable.unwrap_error())
+
+            if r_suitable.unwrap() is True:
+                suitable_flavors.append(flavor)
 
         gluetool.log.log_dict(logger.debug, 'suitable flavors', suitable_flavors)
 
@@ -1114,8 +1393,8 @@ class PoolDriver(gluetool.log.LoggerMixin):
 
         gluetool.log.log_dict(logger.debug, 'sorted suitable flavors', sorted_suitable_flavors)
 
-        gluetool.log.log_dict(logger.debug, 'environment', environment.serialize_to_json())
-        gluetool.log.log_blob(logger.debug, 'constraints', constraints.format())  # noqa: FS002
+        gluetool.log.log_dict(logger.debug, 'environment', environment.serialize())
+        log_dict_yaml(logger.debug, 'constraints', constraints.serialize())
 
         return Ok(sorted_suitable_flavors)
 
@@ -1131,12 +1410,12 @@ class PoolDriver(gluetool.log.LoggerMixin):
         scrubbed_details: Any = {}
 
         if flavor is not None:
-            details['flavor'] = flavor.serialize_to_json()
-            scrubbed_details['flavor'] = flavor.serialize_to_json_scrubbed()
+            details['flavor'] = flavor.serialize()
+            scrubbed_details['flavor'] = flavor.serialize_scrubbed()
 
         if image is not None:
-            details['image'] = image.serialize_to_json()
-            scrubbed_details['image'] = image.serialize_to_json_scrubbed()
+            details['image'] = image.serialize()
+            scrubbed_details['image'] = image.serialize_scrubbed()
 
         log_dict_yaml(logger.info, 'provisioning from', details)
 
@@ -1338,15 +1617,49 @@ class PoolDriver(gluetool.log.LoggerMixin):
         """
         raise NotImplementedError()
 
+    def get_guest_log_updater(self, guest_log: GuestLog) -> Optional[GuestLogUpdaterType]:
+        updater_name = self.guest_log_updaters.get((
+            self.drivername,
+            guest_log.logname,
+            GuestLogContentType(guest_log.contenttype)
+        ))
+
+        if updater_name is None:
+            return None
+
+        return cast(GuestLogUpdaterType, getattr(self, updater_name, None))
+
     def update_guest_log(
         self,
+        logger: gluetool.log.ContextAdapter,
         guest_request: GuestRequest,
         guest_log: GuestLog
     ) -> Result[GuestLogUpdateProgress, Failure]:
-        # cannot provide guest log: functionality not supported for this cloud driver
-        return Ok(GuestLogUpdateProgress(
-            state=GuestLogState.ERROR
-        ))
+        """
+        Call driver-specific code to update a given guest log.
+
+        Guest log updater can be any method accepting guest request and log objcts and decorated
+        with :py:attr:`PoolDriver.guest_log_updater`:
+
+        .. code-block:: python
+
+           @PoolDriver.guest_log_updater('driver name', 'console', GuestLogContentType.BLOB)
+           def _update_guest_log_console_blob(
+               self,
+               guest_request: GuestRequest,
+               guest_log: GuestLog
+           ) -> Result[GuestLogUpdateProgress, Failure]:
+               ...
+        """
+
+        updater = self.get_guest_log_updater(guest_log)
+
+        if updater is None:
+            return Ok(GuestLogUpdateProgress(
+                state=GuestLogState.ERROR
+            ))
+
+        return updater(logger, guest_request, guest_log)
 
     def adjust_capabilities(self, capabilities: PoolCapabilities) -> Result[PoolCapabilities, Failure]:
         """
@@ -1361,7 +1674,7 @@ class PoolDriver(gluetool.log.LoggerMixin):
     def capabilities(self) -> Result[PoolCapabilities, Failure]:
         capabilities = PoolCapabilities()
 
-        capabilities_config = self.pool_config.get('capabilities')
+        capabilities_config = cast(ConfigCapabilitiesType, self.pool_config.get('capabilities'))
 
         if not capabilities_config:
             return self.adjust_capabilities(capabilities)
@@ -1388,12 +1701,12 @@ class PoolDriver(gluetool.log.LoggerMixin):
 
         if 'supports-snapshots' in capabilities_config:
             capabilities.supports_snapshots = gluetool.utils.normalize_bool_option(
-                cast(str, capabilities_config['supports-snapshots'])
+                capabilities_config['supports-snapshots']
             )
 
         if 'supports-spot-instances' in capabilities_config:
             capabilities.supports_spot_instances = gluetool.utils.normalize_bool_option(
-                cast(str, capabilities_config['supports-spot-instances'])
+                capabilities_config['supports-spot-instances']
             )
 
         r_adjusted_capabilities = self.adjust_capabilities(capabilities)
@@ -1404,7 +1717,7 @@ class PoolDriver(gluetool.log.LoggerMixin):
         capabilities = r_adjusted_capabilities.unwrap()
 
         if 'disable-guest-logs' in capabilities_config:
-            for log_spec in cast(List[ConfigCapabilitiesDisableGuestLog], capabilities_config['disable-guest-logs']):
+            for log_spec in capabilities_config['disable-guest-logs']:
                 disabled_log = (log_spec['log-name'], GuestLogContentType(log_spec['content-type']))
 
                 capabilities.supported_guest_logs = [
@@ -1714,7 +2027,7 @@ class PoolDriver(gluetool.log.LoggerMixin):
         Retrieve pool image info for an image of a given name.
         """
 
-        return get_cached_item(CACHE.get(), self.image_info_cache_key, imagename, self.image_info_class)
+        return get_cached_set_item(CACHE.get(), self.image_info_cache_key, imagename, self.image_info_class)
 
     def fetch_pool_flavor_info(self) -> Result[List[Flavor], Failure]:
         """
@@ -1858,21 +2171,21 @@ class PoolDriver(gluetool.log.LoggerMixin):
            is unknown.
         """
 
-        return get_cached_item(CACHE.get(), self.flavor_info_cache_key, flavorname, self.flavor_info_class)
+        return get_cached_set_item(CACHE.get(), self.flavor_info_cache_key, flavorname, self.flavor_info_class)
 
     def get_cached_pool_image_infos(self) -> Result[List[PoolImageInfo], Failure]:
         """
         Retrieve pool image info for all known images.
         """
 
-        return get_cached_items_as_list(CACHE.get(), self.image_info_cache_key, self.image_info_class)
+        return get_cached_set_as_list(CACHE.get(), self.image_info_cache_key, self.image_info_class)
 
     def get_cached_pool_flavor_infos(self) -> Result[List[Flavor], Failure]:
         """
         Retrieve all flavor info known to the pool.
         """
 
-        return get_cached_items_as_list(CACHE.get(), self.flavor_info_cache_key, self.flavor_info_class)
+        return get_cached_set_as_list(CACHE.get(), self.flavor_info_cache_key, self.flavor_info_class)
 
 
 def vm_info_to_ip(output: Any, key: str, regex: Optional[Pattern[str]] = None) -> Result[Optional[str], Failure]:
@@ -1902,7 +2215,8 @@ def run_cli_tool(
     env: Optional[Dict[str, str]] = None,
     # for CLI calls metrics
     poolname: Optional[str] = None,
-    commandname: Optional[str] = None
+    commandname: Optional[str] = None,
+    cause_extractor: Optional[CLIErrorCauseExtractor] = None
 ) -> Result[CLIOutput, Failure]:
     """
     Run a given command, and return its output.
@@ -1943,7 +2257,8 @@ def run_cli_tool(
                 poolname,
                 commandname,
                 output.exit_code,
-                command_time
+                command_time,
+                cause=cause_extractor(output) if cause_extractor is not None else None
             )
 
         joined_command = command_join(command)
@@ -1960,6 +2275,8 @@ def run_cli_tool(
                 'detected a slow CLI command',
                 command_output=output,
                 scrubbed_command=command_scrubber(command),
+                poolname=poolname,
+                commandname=commandname,
                 time=command_time
             ).handle(logger, label='CLI output')
 
@@ -1984,7 +2301,9 @@ def run_cli_tool(
             'error running CLI command',
             exc,
             command_output=exc.output,
-            scrubbed_command=command_scrubber(command)
+            scrubbed_command=command_scrubber(command),
+            poolname=poolname,
+            commandname=commandname
         ))
 
     else:
@@ -1995,7 +2314,9 @@ def run_cli_tool(
             return Error(Failure(
                 'CLI did not emit any output',
                 command_output=output,
-                scrubbed_command=command_scrubber(command)
+                scrubbed_command=command_scrubber(command),
+                poolname=poolname,
+                commandname=commandname
             ))
 
         output_stdout = ''
@@ -2014,7 +2335,9 @@ def run_cli_tool(
             return Error(Failure(
                 'CLI did not emit any output, cannot treat as JSON',
                 command_output=output,
-                scrubbed_command=command_scrubber(command)
+                scrubbed_command=command_scrubber(command),
+                poolname=poolname,
+                commandname=commandname
             ))
 
         try:
@@ -2025,29 +2348,14 @@ def run_cli_tool(
                 'failed to convert string to JSON',
                 exc=exc,
                 command_output=output,
-                scrubbed_command=command_scrubber(command)
+                scrubbed_command=command_scrubber(command),
+                poolname=poolname,
+                commandname=commandname
             ))
 
     return Ok(CLIOutput(output, output_stdout))
 
 
-def test_cli_error(failure: Failure, error_pattern: Pattern[str]) -> bool:
-    if 'command_output' not in failure.details:
-        return False
-
-    os_output = cast(gluetool.utils.ProcessOutput, failure.details['command_output'])
-
-    if not os_output.stderr:
-        return False
-
-    stderr = cast(bytes, os_output.stderr).decode('utf-8')
-
-    if error_pattern.match(stderr):
-        return True
-
-    return False
-
-
 def run_remote(
     logger: gluetool.log.ContextAdapter,
     guest_request: GuestRequest,
@@ -2057,7 +2365,8 @@ def run_remote(
     ssh_timeout: int,
     # for CLI calls metrics
     poolname: Optional[str] = None,
-    commandname: Optional[str] = None
+    commandname: Optional[str] = None,
+    cause_extractor: Optional[CLIErrorCauseExtractor] = None
 ) -> Result[CLIOutput, Failure]:
     if guest_request.address is None:
         return Error(Failure('cannot connect to unknown remote address'))
@@ -2079,7 +2388,8 @@ def run_remote(
                 command_join(command)
             ],
             poolname=poolname,
-            commandname=commandname
+            commandname=commandname,
+            cause_extractor=cause_extractor
         )
 
 
@@ -2093,7 +2403,8 @@ def copy_to_remote(
     ssh_timeout: int,
     # for CLI calls metrics
     poolname: Optional[str] = None,
-    commandname: Optional[str] = None
+    commandname: Optional[str] = None,
+    cause_extractor: Optional[CLIErrorCauseExtractor] = None
 ) -> Result[CLIOutput, Failure]:
     if guest_request.address is None:
         return Error(Failure('cannot connect to unknown remote address'))
@@ -2112,7 +2423,8 @@ def copy_to_remote(
                 f'{guest_request.ssh_username}@{guest_request.address}:{dst}',
             ],
             poolname=poolname,
-            commandname=commandname
+            commandname=commandname,
+            cause_extractor=cause_extractor
         )
 
 
@@ -2126,7 +2438,8 @@ def copy_from_remote(
     ssh_timeout: int,
     # for CLI calls metrics
     poolname: Optional[str] = None,
-    commandname: Optional[str] = None
+    commandname: Optional[str] = None,
+    cause_extractor: Optional[CLIErrorCauseExtractor] = None
 ) -> Result[CLIOutput, Failure]:
     if guest_request.address is None:
         return Error(Failure('cannot connect to unknown remote address'))
@@ -2145,7 +2458,8 @@ def copy_from_remote(
                 dst
             ],
             poolname=poolname,
-            commandname=commandname
+            commandname=commandname,
+            cause_extractor=cause_extractor
         )
 
 
@@ -2157,7 +2471,8 @@ def ping_shell_remote(
     ssh_timeout: int,
     # for CLI calls metrics
     poolname: Optional[str] = None,
-    commandname: Optional[str] = None
+    commandname: Optional[str] = None,
+    cause_extractor: Optional[CLIErrorCauseExtractor] = None
 ) -> Result[bool, Failure]:
     """
     Try to run a simple ``echo`` command on a given guest, and verify its output.
@@ -2177,7 +2492,8 @@ def ping_shell_remote(
         key=key,
         ssh_timeout=ssh_timeout,
         poolname=poolname,
-        commandname=commandname
+        commandname=commandname,
+        cause_extractor=cause_extractor
     )
 
     if r_ssh.is_error:
