@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import dataclasses
 import threading
 from typing import Any, Dict, List, Optional, Union, cast
@@ -6,13 +9,38 @@ import gluetool.log
 import sqlalchemy.orm.session
 from gluetool.result import Error, Ok, Result
 
+from tft.artemis.drivers.aws import awscli_error_cause_extractor
+
 from .. import Failure, JSONType, log_dict_yaml
 from ..db import GuestRequest, SnapshotRequest
+from ..environment import FlavorBoot
+from ..knobs import Knob
 from ..metrics import ResourceType
 from . import KNOB_UPDATE_GUEST_REQUEST_TICK, HookImageInfoMapper, PoolCapabilities, PoolData, PoolDriver, \
     PoolImageInfo, PoolImageSSHInfo, PoolResourcesIDs, ProvisioningProgress, ProvisioningState, \
     SerializedPoolResourcesIDs, create_tempfile, run_cli_tool, vm_info_to_ip
 
+KNOB_ENVIRONMENT_TO_IMAGE_MAPPING_FILEPATH: Knob[str] = Knob(
+    'azure.mapping.environment-to-image.pattern-map.filepath',
+    'Path to a pattern map file with environment to image mapping.',
+    has_db=False,
+    per_pool=True,
+    envvar='ARTEMIS_AZURE_ENVIRONMENT_TO_IMAGE_MAPPING_FILEPATH',
+    cast_from_str=str,
+    default='artemis-image-map-azure.yaml'
+)
+
+KNOB_ENVIRONMENT_TO_IMAGE_MAPPING_NEEDLE: Knob[str] = Knob(
+    'azure.mapping.environment-to-image.pattern-map.needle',
+    'A pattern for needle to match in environment to image mapping file.',
+    has_db=False,
+    per_pool=True,
+    envvar='ARTEMIS_AZURE_ENVIRONMENT_TO_IMAGE_MAPPING_NEEDLE',
+    cast_from_str=str,
+    default='{{ os.compose }}'
+)
+
+
 AZURE_RESOURCE_TYPE: Dict[str, ResourceType] = {
     'Microsoft.Compute/virtualMachines': ResourceType.VIRTUAL_MACHINE,
     'Microsoft.Network/virtualNetworks': ResourceType.VIRTUAL_NETWORK,
@@ -36,13 +64,17 @@ class AzurePoolResourcesIDs(PoolResourcesIDs):
 
 
 class AzureDriver(PoolDriver):
+    drivername = 'azure'
+
+    pool_data_class = AzurePoolData
+
     def __init__(
         self,
         logger: gluetool.log.ContextAdapter,
         poolname: str,
         pool_config: Dict[str, Any],
     ) -> None:
-        super(AzureDriver, self).__init__(logger, poolname, pool_config)
+        super().__init__(logger, poolname, pool_config)
 
     @property
     def image_info_mapper(self) -> HookImageInfoMapper[PoolImageInfo]:
@@ -78,11 +110,15 @@ class AzureDriver(PoolDriver):
         )
 
         if r_images_show.is_error:
-            return Error(r_images_show.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch image information',
+                r_images_show.unwrap_error()
+            ))
 
         return Ok(PoolImageInfo(
             name=imagename,
             id=imagename,
+            boot=FlavorBoot(),
             ssh=PoolImageSSHInfo()
         ))
 
@@ -96,7 +132,7 @@ class AzureDriver(PoolDriver):
 
         # delete vm first, resources second
 
-        resource_ids = AzurePoolResourcesIDs.unserialize(raw_resource_ids)
+        resource_ids = AzurePoolResourcesIDs.unserialize_from_json(raw_resource_ids)
 
         def _delete_resource(res_id: str) -> Any:
             options = ['resource', 'delete', '--ids', res_id]
@@ -132,7 +168,7 @@ class AzureDriver(PoolDriver):
         the given environment.
         """
 
-        r_answer = super(AzureDriver, self).can_acquire(logger, session, guest_request)
+        r_answer = super().can_acquire(logger, session, guest_request)
 
         if r_answer.is_error:
             return Error(r_answer.unwrap_error())
@@ -336,7 +372,8 @@ class AzureDriver(PoolDriver):
             json_output=json_format,
             command_scrubber=lambda cmd: (['azure'] + options),
             poolname=self.poolname,
-            commandname=commandname
+            commandname=commandname,
+            cause_extractor=awscli_error_cause_extractor
         )
 
         if r_run.is_error:
@@ -372,7 +409,10 @@ class AzureDriver(PoolDriver):
             ], commandname='az.login')
 
             if login_output.is_error:
-                return Error(login_output.unwrap_error())
+                return Error(Failure.from_failure(
+                    'failed to log into tenant',
+                    login_output.unwrap_error()
+                ))
 
         return Ok(None)
 
@@ -388,7 +428,10 @@ class AzureDriver(PoolDriver):
         ], commandname='az.vm-show')
 
         if r_output.is_error:
-            return Error(r_output.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch instance information',
+                r_output.unwrap_error()
+            ))
         return Ok(r_output.unwrap())
 
     def _do_acquire_guest(
@@ -495,3 +538,6 @@ class AzureDriver(PoolDriver):
             delay_update=r_delay.unwrap(),
             ssh_info=image.ssh
         ))
+
+
+PoolDriver._drivers_registry['azure'] = AzureDriver
