@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import dataclasses
 import os
 import re
@@ -10,17 +13,20 @@ import gluetool.log
 import gluetool.utils
 import pint
 import sqlalchemy.orm.session
-from gluetool.log import log_xml
+from gluetool.log import ContextAdapter, log_xml
 from gluetool.result import Error, Ok, Result
+from typing_extensions import TypedDict
 
-from .. import Failure, log_dict_yaml
-from ..db import GuestRequest
-from ..environment import And, Constraint, ConstraintBase, Environment, Operator, Or
+from .. import Failure, SerializableContainer, log_dict_yaml, process_output_to_str
+from ..cache import get_cached_set, refresh_cached_set
+from ..context import CACHE
+from ..db import GuestLog, GuestLogContentType, GuestLogState, GuestRequest
+from ..environment import And, Constraint, ConstraintBase, Environment, FlavorBoot, Operator, Or
 from ..knobs import Knob
 from ..metrics import PoolMetrics, PoolResourcesMetrics, ResourceType
-from . import KNOB_UPDATE_GUEST_REQUEST_TICK, CLIOutput, HookImageInfoMapper, PoolData, PoolDriver, PoolImageInfo, \
-    PoolImageSSHInfo, PoolResourcesIDs, ProvisioningProgress, ProvisioningState, SerializedPoolResourcesIDs, \
-    create_tempfile, run_cli_tool, test_cli_error
+from . import KNOB_UPDATE_GUEST_REQUEST_TICK, CLIErrorCauses, CLIOutput, GuestLogUpdateProgress, HookImageInfoMapper, \
+    PoolCapabilities, PoolData, PoolDriver, PoolImageInfo, PoolImageSSHInfo, PoolResourcesIDs, ProvisioningProgress, \
+    ProvisioningState, SerializedPoolResourcesIDs, create_tempfile, guest_log_updater, run_cli_tool
 
 NodeRefType = Any
 
@@ -34,8 +40,53 @@ KNOB_RESERVATION_DURATION: Knob[int] = Knob(
     default=86400
 )
 
+KNOB_ENVIRONMENT_TO_IMAGE_MAPPING_FILEPATH: Knob[str] = Knob(
+    'beaker.mapping.environment-to-image.pattern-map.filepath',
+    'Path to a pattern map file with environment to image mapping.',
+    has_db=False,
+    per_pool=True,
+    envvar='ARTEMIS_BEAKER_ENVIRONMENT_TO_IMAGE_MAPPING_FILEPATH',
+    cast_from_str=str,
+    default='artemis-image-map-beaker.yaml'
+)
 
-NO_DISTRO_MATCHES_RECIPE_ERROR_PATTEN = re.compile(r'^Exception: .+:No distro tree matches Recipe:')
+KNOB_ENVIRONMENT_TO_IMAGE_MAPPING_NEEDLE: Knob[str] = Knob(
+    'beaker.mapping.environment-to-image.pattern-map.needle',
+    'A pattern for needle to match in environment to image mapping file.',
+    has_db=False,
+    per_pool=True,
+    envvar='ARTEMIS_BEAKER_ENVIRONMENT_TO_IMAGE_MAPPING_NEEDLE',
+    cast_from_str=str,
+    default='{{ os.compose }}'
+)
+
+
+class BkrErrorCauses(CLIErrorCauses):
+    NONE = 'none'
+    NO_DISTRO_MATCHES_RECIPE = 'no-distro-matches-recipe'
+
+
+CLI_ERROR_PATTERNS = {
+    BkrErrorCauses.NO_DISTRO_MATCHES_RECIPE: re.compile(r'^Exception: .+:No distro tree matches Recipe:')
+}
+
+
+def bkr_error_cause_extractor(output: gluetool.utils.ProcessOutput) -> BkrErrorCauses:
+    if output.exit_code == 0:
+        return BkrErrorCauses.NONE
+
+    stderr = process_output_to_str(output, stream='stderr')
+
+    if stderr is None:
+        return BkrErrorCauses.NONE
+
+    for cause, pattern in CLI_ERROR_PATTERNS.items():
+        if not pattern.match(stderr):
+            continue
+
+        return cause
+
+    return BkrErrorCauses.NONE
 
 
 @dataclasses.dataclass
@@ -48,6 +99,23 @@ class BeakerPoolResourcesIDs(PoolResourcesIDs):
     job_id: Optional[str] = None
 
 
+@dataclasses.dataclass(repr=False)
+class AvoidGroupHostnames(SerializableContainer):
+    groupname: str
+    hostnames: List[str] = dataclasses.field(default_factory=list)
+
+
+class ConstraintTranslationConfigType(TypedDict):
+    operator: str
+    value: str
+    element: str
+
+
+class BootMethodConfigType(TypedDict):
+    key: str
+    translations: List[ConstraintTranslationConfigType]
+
+
 #: Mapping of operators to their Beaker representation. :py:attr:`Operator.MATCH` is missing on purpose: it is
 #: intercepted in :py:func:`operator_to_beaker_op`.
 OPERATOR_SIGN_TO_OPERATOR = {
@@ -76,7 +144,10 @@ def operator_to_beaker_op(operator: Operator, value: str) -> Tuple[str, str]:
     return 'like', value.replace('.*', '%').replace('.+', '%')
 
 
-def constraint_to_beaker_filter(constraint: ConstraintBase) -> Result[bs4.BeautifulSoup, Failure]:
+def constraint_to_beaker_filter(
+    constraint: ConstraintBase,
+    pool: 'BeakerDriver'
+) -> Result[bs4.BeautifulSoup, Failure]:
     """
     Convert a given constraint to XML tree representing Beaker filter compatible with Beaker's ``hostRequires``
     element.
@@ -86,7 +157,7 @@ def constraint_to_beaker_filter(constraint: ConstraintBase) -> Result[bs4.Beauti
         grouping_and = _new_tag('and')
 
         for child_constraint in constraint.constraints:
-            r_child_element = constraint_to_beaker_filter(child_constraint)
+            r_child_element = constraint_to_beaker_filter(child_constraint, pool)
 
             if r_child_element.is_error:
                 return Error(r_child_element.unwrap_error())
@@ -99,7 +170,7 @@ def constraint_to_beaker_filter(constraint: ConstraintBase) -> Result[bs4.Beauti
         grouping_or = _new_tag('or')
 
         for child_constraint in constraint.constraints:
-            r_child_element = constraint_to_beaker_filter(child_constraint)
+            r_child_element = constraint_to_beaker_filter(child_constraint, pool)
 
             if r_child_element.is_error:
                 return Error(r_child_element.unwrap_error())
@@ -109,44 +180,80 @@ def constraint_to_beaker_filter(constraint: ConstraintBase) -> Result[bs4.Beauti
         return Ok(grouping_or)
 
     constraint = cast(Constraint, constraint)
+    constraint_name = constraint.expand_name()
 
-    if constraint.name.startswith('cpu.'):
+    if constraint_name.property == 'boot':
+        if constraint_name.child_property == 'method':
+            boot_method_config = cast(
+                BootMethodConfigType,
+                pool.pool_config.get('hw-constraints', {}).get('boot', {}).get('method', {})
+            )
+
+            for translation in boot_method_config.get('translations', []):
+                if translation['operator'] != constraint.operator.value or translation['value'] != constraint.value:
+                    continue
+
+                try:
+                    return Ok(bs4.BeautifulSoup(translation['element'], 'xml'))
+
+                except Exception as exc:
+                    return Error(Failure.from_exc(
+                        'failed to parse XML',
+                        exc,
+                        source=translation['element']
+                    ))
+
+            return Error(Failure(
+                'constraint not supported by driver',
+                constraint=repr(constraint)
+            ))
+
+    if constraint_name.property == 'cpu':
         cpu = _new_tag('cpu')
 
-        if constraint.name == 'cpu.cores':
+        if constraint_name.child_property == 'cores':
             op, value = operator_to_beaker_op(constraint.operator, str(constraint.value))
 
-            processors = _new_tag('processors', op=op, value=value)
+            cores = _new_tag('cores', op=op, value=value)
 
-            cpu.append(processors)
+            cpu.append(cores)
 
-        elif constraint.name == 'cpu.model':
+        elif constraint_name.child_property == 'family':
             op, value = operator_to_beaker_op(constraint.operator, str(constraint.value))
 
-            processors = _new_tag('model', op=op, value=value)
+            family = _new_tag('family', op=op, value=value)
 
-            cpu.append(processors)
+            cpu.append(family)
 
-        elif constraint.name == 'cpu.model_name':
+        elif constraint_name.child_property == 'model':
             op, value = operator_to_beaker_op(constraint.operator, str(constraint.value))
 
-            processors = _new_tag('model_name', op=op, value=value)
+            model = _new_tag('model', op=op, value=value)
 
-            cpu.append(processors)
+            cpu.append(model)
+
+        elif constraint_name.child_property == 'model_name':
+            op, value = operator_to_beaker_op(constraint.operator, str(constraint.value))
+
+            model_name = _new_tag('model_name', op=op, value=value)
+
+            cpu.append(model_name)
 
         else:
             return Error(Failure(
                 'contraint not supported by driver',
-                constraint=constraint.format()  # noqa: FS002
+                constraint=repr(constraint)
             ))
 
         return Ok(cpu)
 
-    # TODO: support more than one disk - but this is a stand-alone feature spanning more than this driver.
-    if constraint.name.startswith('disk[0].'):
+    if constraint_name.property == 'disk':
+        if constraint_name.child_property in ('is_expansion', 'min_size', 'max_size', 'length', 'expanded_length'):
+            return Ok(_new_tag('or'))
+
         disk = _new_tag('disk')
 
-        if constraint.name == 'disk[0].size':
+        if constraint_name.child_property == 'size':
             # `disk.size` is represented as quantity, for Beaker XML we need to convert to bytes, integer.
             op, value = operator_to_beaker_op(
                 constraint.operator,
@@ -160,12 +267,13 @@ def constraint_to_beaker_filter(constraint: ConstraintBase) -> Result[bs4.Beauti
         else:
             return Error(Failure(
                 'contraint not supported by driver',
-                constraint=constraint.format()  # noqa: FS002
+                constraint=repr(constraint),
+                constraint_name=constraint.name
             ))
 
         return Ok(disk)
 
-    if constraint.name == 'arch':
+    if constraint_name.property == 'arch':
         op, value = operator_to_beaker_op(constraint.operator, str(constraint.value))
 
         system = _new_tag('system')
@@ -175,7 +283,7 @@ def constraint_to_beaker_filter(constraint: ConstraintBase) -> Result[bs4.Beauti
 
         return Ok(system)
 
-    if constraint.name == 'memory':
+    if constraint_name.property == 'memory':
         # `memory` is represented as quantity, for Beaker XML we need to convert to mibibytes, integer.
         op, value = operator_to_beaker_op(
             constraint.operator,
@@ -189,13 +297,50 @@ def constraint_to_beaker_filter(constraint: ConstraintBase) -> Result[bs4.Beauti
 
         return Ok(system)
 
+    if constraint_name.property == 'hostname':
+        op, value = operator_to_beaker_op(
+            constraint.operator,
+            str(constraint.value)
+        )
+
+        hostname = _new_tag('hostname', op=op, value=value)
+
+        return Ok(hostname)
+
     return Error(Failure(
         'contraint not supported by driver',
-        constraint=constraint.format()  # noqa: FS002
+        constraint=repr(constraint),
+        constraint_name=constraint.name
     ))
 
 
-def environment_to_beaker_filter(environment: Environment) -> Result[bs4.BeautifulSoup, Failure]:
+def _prune_beaker_filter(tree: bs4.BeautifulSoup) -> Result[bs4.BeautifulSoup, Failure]:
+    # Conversion process produces empty `and` and `or` tags, thanks to how Beaker deals with disks without flavors.
+    # The following code is a crude attempt to get rid of some of them (there may be some left, empty `or` in `and`
+    # in `or`) would keep the last `or` since we run just two swipes. But that's good enough for now.
+    def _remove_empty(tag_name: str) -> None:
+        for el in tree.find_all(tag_name):
+            if len(el.contents) == 0:
+                el.extract()
+
+    _remove_empty('or')
+    _remove_empty('and')
+
+    def _remove_singles(tag_name: str) -> None:
+        for el in tree.find_all(tag_name):
+            if len(el.contents) == 1:
+                el.replace_with(el.contents[0])
+
+    _remove_singles('or')
+    _remove_singles('and')
+
+    return Ok(tree)
+
+
+def environment_to_beaker_filter(
+    environment: Environment,
+    pool: 'BeakerDriver'
+) -> Result[bs4.BeautifulSoup, Failure]:
     """
     Convert a given environment to Beaker XML tree representing Beaker filter compatible with Beaker's ``hostRequires``
     element.
@@ -205,6 +350,9 @@ def environment_to_beaker_filter(environment: Environment) -> Result[bs4.Beautif
        Converts the `environment`, not just the constraints: in our world, ``arch`` stands separated from
        the constraints while in Beaker, ``arch`` is part of the XML filter subtree. Therefore if there are no
        constraints, this helper emits a XML filter based on architecture alone.
+
+    :param environment: environment as a source of constraints.
+    :returns: a constraint representing the environment.
     """
 
     r_constraints = environment.get_hw_constraints()
@@ -215,22 +363,204 @@ def environment_to_beaker_filter(environment: Environment) -> Result[bs4.Beautif
     constraints = r_constraints.unwrap()
 
     if constraints is None:
-        return constraint_to_beaker_filter(Constraint.from_arch(environment.hw.arch))
+        r_beaker_filter = constraint_to_beaker_filter(Constraint.from_arch(environment.hw.arch), pool)
 
-    return constraint_to_beaker_filter(And([
-        Constraint.from_arch(environment.hw.arch),
-        constraints
-    ]))
+    else:
+        r_beaker_filter = constraint_to_beaker_filter(And([
+            Constraint.from_arch(environment.hw.arch),
+            constraints
+        ]), pool)
+
+    if r_beaker_filter.is_error:
+        return r_beaker_filter
+
+    return _prune_beaker_filter(r_beaker_filter.unwrap())
+
+
+def groups_to_beaker_filter(avoid_groups: List[str]) -> Result[bs4.BeautifulSoup, Failure]:
+    """
+    Convert given lists of groups to Beaker XML tree representing Beaker filter compatible with Beaker's
+    ``hostRequires`` element.
+
+    For each ``group`` from ``avoid_groups`` list, a following element is created:
+
+    .. code-block:: xml
+
+       <group op="!=" value="$group"/>
+
+    :param avoid_groups: list of Beaker groups to filter out when provisioning.
+    :returns: a Beaker filter representing groups to avoid.
+    """
+
+    # This is a container for all per-tag group elements. When called with an empty list, this would result
+    # in an empty <and/> element being returned - but that's OK, _prune_beaker_filter() can deal with such
+    # elements, and it simplifies handling of return values, no pesky `if is None` tests.
+    container = _new_tag('and')
+
+    for group in avoid_groups:
+        container.append(_new_tag('group', op='!=', value=group))
+
+    return Ok(container)
+
+
+def hostnames_to_beaker_filter(avoid_hostnames: List[str]) -> Result[bs4.BeautifulSoup, Failure]:
+    """
+    Convert given lists of hostnames to Beaker XML tree representing Beaker filter compatible with Beaker's
+    ``hostRequires`` element.
+
+    For each ``hostname`` from ``avoid_hostnames`` list, a following element is created:
+
+    .. code-block:: xml
+
+       <hostname op="!=" value="$hostname"/>
+
+    :param avoid_hostnames: list of Beaker hostnames to filter out when provisioning.
+    :returns: a Beaker filter representing hostnamesto avoid.
+    """
+
+    # This is a container for all per-tag hostname elements. When called with an empty list, this would result
+    # in an empty <and/> element being returned - but that's OK, _prune_beaker_filter() can deal with such
+    # elements, and it simplifies handling of return values, no pesky `if is None` tests.
+    container = _new_tag('and')
+
+    for hostname in avoid_hostnames:
+        container.append(_new_tag('hostname', op='!=', value=hostname))
+
+    return Ok(container)
+
+
+def merge_beaker_filters(filters: List[bs4.BeautifulSoup]) -> Result[bs4.BeautifulSoup, Failure]:
+    """
+    Merge given Beaker filters into a single filter.
+
+    Each filter must be a single element, but may have child elements. Filters would be merged into
+    a single tree in which all filters must be satisfied for the final filter to be satisfied.
+
+    :param filters: filters to merge.
+    :returns: a Beaker filter taking all given filters into account.
+    """
+
+    # When called with an empty list, this would result in an empty <and/> element being returned - but that's OK,
+    # _prune_beaker_filter() can deal with such elements, and it simplifies handling of return values, no pesky
+    # `if is None` tests.
+    if not filters:
+        return Ok(_new_tag('and'))
+
+    container = filters.pop(0)
+
+    while filters:
+        next_filter = filters.pop(0)
+
+        # When the container is already an <and/> element, we can simply add the next filter to container's
+        # children.
+        if container.name == 'and':
+            # Children of two <and/> elements can be easily merged, avoiding a nested <and/>.
+            if next_filter.name == 'and':
+                # TODO: for some reason, without a list only the first child is yielded from .children :/
+                for child in list(next_filter.children):
+                    container.append(child)
+
+            else:
+                container.append(next_filter)
+
+        # For any other element, we need to create a new container, <and/>, and add our current container to it
+        # as well as the next filter. That way, both branches needs to be satisfied for the final filter to be
+        # happy.
+        else:
+            new_container = _new_tag('and')
+
+            new_container.append(container)
+            new_container.append(next_filter)
+
+            container = new_container
+
+    return _prune_beaker_filter(container)
+
+
+def create_beaker_filter(
+    environment: Environment,
+    pool: 'BeakerDriver',
+    avoid_groups: List[str],
+    avoid_hostnames: List[str]
+) -> Result[Optional[bs4.BeautifulSoup], Failure]:
+    """
+    From given inputs, create a Beaker filter.
+
+    :param environment: environment as a source of constraints.
+    :param avoid_groups: list of Beaker groups to filter out when provisioning.
+    :param avoid_hostnames: list of Beaker hostnames to filter out when provisioning.
+    :returns: a Beaker filter taking all given inputs into account.
+    """
+
+    beaker_filters: List[bs4.BeautifulSoup] = []
+
+    if environment.has_hw_constraints:
+        r_beaker_filter = environment_to_beaker_filter(environment, pool)
+
+        if r_beaker_filter.is_error:
+            return Error(r_beaker_filter.unwrap_error())
+
+        beaker_filters.append(r_beaker_filter.unwrap())
+
+    if avoid_groups:
+        r_beaker_filter = groups_to_beaker_filter(avoid_groups)
+
+        if r_beaker_filter.is_error:
+            return Error(r_beaker_filter.unwrap_error())
+
+        beaker_filters.append(r_beaker_filter.unwrap())
+
+    if avoid_hostnames:
+        r_beaker_filter = hostnames_to_beaker_filter(avoid_hostnames)
+
+        if r_beaker_filter.is_error:
+            return Error(r_beaker_filter.unwrap_error())
+
+        beaker_filters.append(r_beaker_filter.unwrap())
+
+    if not beaker_filters:
+        return Ok(None)
+
+    r_beaker_filter = merge_beaker_filters(beaker_filters)
+
+    if r_beaker_filter.is_error:
+        return Error(r_beaker_filter.unwrap_error())
+
+    return _prune_beaker_filter(r_beaker_filter.unwrap())
+
+
+def _create_wow_options(
+    logger: gluetool.log.ContextAdapter,
+    guest_request: GuestRequest,
+    distro: PoolImageInfo
+) -> Result[List[str], Failure]:
+    return Ok([
+        'workflow-simple',
+        '--dry-run',
+        '--prettyxml',
+        '--distro', distro.id,
+        '--arch', guest_request.environment.hw.arch,
+        '--task', '/distribution/dummy',
+        '--reserve',
+        '--reserve-duration', str(KNOB_RESERVATION_DURATION.value)
+    ])
 
 
 class BeakerDriver(PoolDriver):
+    drivername = 'beaker'
+
+    pool_data_class = BeakerPoolData
+
+    #: Template for a cache key holding avoid groups hostnames.
+    POOL_AVOID_GROUPS_HOSTNAMES_CACHE_KEY = 'pool.{}.avoid-groups.hostnames'
+
     def __init__(
         self,
         logger: gluetool.log.ContextAdapter,
         poolname: str,
         pool_config: Dict[str, Any]
     ) -> None:
-        super(BeakerDriver, self).__init__(logger, poolname, pool_config)
+        super().__init__(logger, poolname, pool_config)
 
         # Prepare `bkr` command with options we can already deduce.
         self._bkr_command = [
@@ -243,10 +573,38 @@ class BeakerDriver(PoolDriver):
                 '--password', self.pool_config['password']
             ]
 
+        self.avoid_groups_hostnames_cache_key = self.POOL_AVOID_GROUPS_HOSTNAMES_CACHE_KEY.format(self.poolname)  # noqa: FS002,E501
+
+    def adjust_capabilities(self, capabilities: PoolCapabilities) -> Result[PoolCapabilities, Failure]:
+        capabilities.supports_hostnames = True
+        capabilities.supported_guest_logs = [
+            ('console', GuestLogContentType.URL),
+            ('console.log', GuestLogContentType.URL),
+            ('sys.log', GuestLogContentType.URL),
+        ]
+
+        return Ok(capabilities)
+
     @property
     def image_info_mapper(self) -> HookImageInfoMapper[PoolImageInfo]:
         return HookImageInfoMapper(self, 'BEAKER_ENVIRONMENT_TO_IMAGE')
 
+    @property
+    def avoid_groups(self) -> Result[List[str], Failure]:
+        return Ok(self.pool_config.get('avoid-groups', []))
+
+    @property
+    def avoid_hostnames(self) -> Result[List[str], Failure]:
+        r_avoid_hostnames = self.get_avoid_groups_hostnames()
+
+        if r_avoid_hostnames.is_error:
+            return Error(r_avoid_hostnames.unwrap_error())
+
+        return Ok(sum(
+            (group.hostnames for group in r_avoid_hostnames.unwrap().values()),
+            self.pool_config.get('avoid-hostnames', [])
+        ))
+
     def _run_bkr(
         self,
         logger: gluetool.log.ContextAdapter,
@@ -267,7 +625,8 @@ class BeakerDriver(PoolDriver):
             self._bkr_command + options,
             json_output=False,
             poolname=self.poolname,
-            commandname=commandname
+            commandname=commandname,
+            cause_extractor=bkr_error_cause_extractor
         )
 
         if r_run.is_error:
@@ -288,7 +647,7 @@ class BeakerDriver(PoolDriver):
 
         failure.recoverable = False
 
-        PoolMetrics.inc_error(self.poolname, 'no-distro-matches-recipe')
+        PoolMetrics.inc_error(self.poolname, BkrErrorCauses.NO_DISTRO_MATCHES_RECIPE.value)
 
         return Ok(ProvisioningProgress(
             state=ProvisioningState.CANCEL,
@@ -311,13 +670,16 @@ class BeakerDriver(PoolDriver):
         logger: gluetool.log.ContextAdapter,
         raw_resource_ids: SerializedPoolResourcesIDs
     ) -> Result[None, Failure]:
-        resource_ids = BeakerPoolResourcesIDs.unserialize(raw_resource_ids)
+        resource_ids = BeakerPoolResourcesIDs.unserialize_from_json(raw_resource_ids)
 
         if resource_ids.job_id is not None:
             r_output = self._run_bkr(logger, ['job-cancel', resource_ids.job_id], commandname='bkr.job-cancel')
 
             if r_output.is_error:
-                return Error(r_output.unwrap_error())
+                return Error(Failure.from_failure(
+                    'failed to cancel job',
+                    r_output.unwrap_error()
+                ))
 
             self.inc_costs(logger, ResourceType.VIRTUAL_MACHINE, resource_ids.ctime)
 
@@ -339,6 +701,7 @@ class BeakerDriver(PoolDriver):
         return Ok(PoolImageInfo(
             name=imagename,
             id=imagename,
+            boot=FlavorBoot(),
             ssh=PoolImageSSHInfo()
         ))
 
@@ -357,15 +720,27 @@ class BeakerDriver(PoolDriver):
         :returns: :py:class:`result.Result` with job xml, or specification of error.
         """
 
-        beaker_filter: Optional[bs4.BeautifulSoup] = None
+        r_avoid_hostnames = self.avoid_hostnames
 
-        if guest_request.environment.has_hw_constraints:
-            r_beaker_filter = environment_to_beaker_filter(guest_request.environment)
+        if r_avoid_hostnames.is_error:
+            return Error(r_avoid_hostnames.unwrap_error())
 
-            if r_beaker_filter.is_error:
-                return Error(r_beaker_filter.unwrap_error())
+        r_avoid_groups = self.avoid_groups
 
-            beaker_filter = r_beaker_filter.unwrap()
+        if r_avoid_groups.is_error:
+            return Error(r_avoid_groups.unwrap_error())
+
+        r_beaker_filter = create_beaker_filter(
+            guest_request.environment,
+            self,
+            r_avoid_groups.unwrap(),
+            r_avoid_hostnames.unwrap()
+        )
+
+        if r_beaker_filter.is_error:
+            return Error(r_beaker_filter.unwrap_error())
+
+        beaker_filter = r_beaker_filter.unwrap()
 
         r_distro = self.image_info_mapper.map(logger, guest_request)
 
@@ -374,6 +749,11 @@ class BeakerDriver(PoolDriver):
 
         distro = r_distro.unwrap()
 
+        r_wow_options = _create_wow_options(logger, guest_request, distro)
+
+        if r_wow_options.is_error:
+            return Error(r_wow_options.unwrap_error())
+
         self.log_acquisition_attempt(
             logger,
             session,
@@ -381,20 +761,12 @@ class BeakerDriver(PoolDriver):
             image=distro
         )
 
-        options = [
-            'workflow-simple',
-            '--dry-run',
-            '--prettyxml',
-            '--distro', distro.id,
-            '--arch', guest_request.environment.hw.arch,
-            '--task', '/distribution/dummy',
-            '--reserve',
-            '--reserve-duration', str(KNOB_RESERVATION_DURATION.value)
-        ]
-
-        r_workflow_simple = self._run_bkr(logger, options, commandname='bkr.workflow-simple')
+        r_workflow_simple = self._run_bkr(logger, r_wow_options.unwrap(), commandname='bkr.workflow-simple')
         if r_workflow_simple.is_error:
-            return Error(r_workflow_simple.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to create job',
+                r_workflow_simple.unwrap_error()
+            ))
 
         bkr_output = r_workflow_simple.unwrap()
 
@@ -422,7 +794,7 @@ class BeakerDriver(PoolDriver):
                 job=job_xml.prettify()
             ))
 
-        host_requires.append(beaker_filter)
+        list(host_requires)[0].append(beaker_filter)
 
         log_xml(logger.debug, 'job with filter', job_xml)
 
@@ -455,7 +827,10 @@ class BeakerDriver(PoolDriver):
             r_job_submit = self._run_bkr(logger, ['job-submit', job_filepath], commandname='bkr.job-submit')
 
         if r_job_submit.is_error:
-            return Error(r_job_submit.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to submit job',
+                r_job_submit.unwrap_error()
+            ))
 
         bkr_output = r_job_submit.unwrap()
 
@@ -496,30 +871,6 @@ class BeakerDriver(PoolDriver):
 
         return self._submit_job(logger, r_job_xml.unwrap())
 
-    def _reschedule_job(
-        self,
-        logger: gluetool.log.ContextAdapter,
-        session: sqlalchemy.orm.session.Session,
-        guest_request: GuestRequest,
-        job_id: str
-    ) -> Result[str, Failure]:
-        """
-        Reschedule a Beaker job. Cancel the old job with `job_id`, create and
-        submit a new job with `environment` specs and return new `job_id`.
-
-        :param str job_id: Job id that will be rescheduled.
-        :param Environment environment: An environment of a guest.
-        :rtype: result.Result[str, Failure]
-        :returns: :py:class:`result.Result` with job id, or specification of error.
-        """
-
-        r_job_cancel = self._dispatch_resource_cleanup(self.logger, job_id=job_id)
-
-        if r_job_cancel.is_error:
-            return Error(r_job_cancel.unwrap_error())
-
-        return self._create_job(logger, session, guest_request)
-
     def _get_job_results(
         self,
         logger: gluetool.log.ContextAdapter,
@@ -536,7 +887,10 @@ class BeakerDriver(PoolDriver):
         r_results = self._run_bkr(logger, ['job-results', job_id], commandname='bkr.job-results')
 
         if r_results.is_error:
-            return Error(r_results.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch job results',
+                r_results.unwrap_error()
+            ))
 
         bkr_output = r_results.unwrap()
 
@@ -670,29 +1024,15 @@ class BeakerDriver(PoolDriver):
             or (job_status == 'reserved' and job_result == 'warn')  # job failed, needs a bit more time to update status
 
         if job_is_failed:
-            r_reschedule_job = self._reschedule_job(
-                logger,
-                session,
-                guest_request,
-                BeakerPoolData.unserialize(guest_request).job_id
-            )
-
-            if r_reschedule_job.is_error:
-                failure = r_reschedule_job.unwrap_error()
-
-                if test_cli_error(failure, NO_DISTRO_MATCHES_RECIPE_ERROR_PATTEN):
-                    return self._handle_no_distro_matches_recipe_error(
-                        failure,
-                        guest_request,
-                        pool_data=BeakerPoolData.unserialize(guest_request)
-                    )
-
-                return Error(r_reschedule_job.unwrap_error())
-
             return Ok(ProvisioningProgress(
-                state=ProvisioningState.PENDING,
-                pool_data=BeakerPoolData(job_id=r_reschedule_job.unwrap()),
-                delay_update=r_delay.unwrap()
+                state=ProvisioningState.CANCEL,
+                pool_data=BeakerPoolData.unserialize(guest_request),
+                pool_failures=[Failure(
+                    'beaker job failed',
+                    job_result=job_result,
+                    job_status=job_status,
+                    job_results=job_results.prettify()
+                )]
             ))
 
         return Error(Failure(
@@ -718,7 +1058,7 @@ class BeakerDriver(PoolDriver):
         """
 
         # First, check the parent class, maybe its tests already have the answer.
-        r_answer = super(BeakerDriver, self).can_acquire(logger, session, guest_request)
+        r_answer = super().can_acquire(logger, session, guest_request)
 
         if r_answer.is_error:
             return Error(r_answer.unwrap_error())
@@ -743,7 +1083,7 @@ class BeakerDriver(PoolDriver):
         # since `has_hw_constraints` was positive, there should be constraints...
         assert constraints is not None
 
-        r_filter = constraint_to_beaker_filter(constraints)
+        r_filter = constraint_to_beaker_filter(constraints, self)
 
         if r_filter.is_error:
             return Error(r_filter.unwrap_error())
@@ -782,7 +1122,8 @@ class BeakerDriver(PoolDriver):
         if r_create_job.is_error:
             failure = r_create_job.unwrap_error()
 
-            if test_cli_error(failure, NO_DISTRO_MATCHES_RECIPE_ERROR_PATTEN):
+            if failure.command_output \
+               and bkr_error_cause_extractor(failure.command_output) == BkrErrorCauses.NO_DISTRO_MATCHES_RECIPE:
                 return self._handle_no_distro_matches_recipe_error(failure, guest_request)
 
             return Error(r_create_job.unwrap_error())
@@ -827,7 +1168,7 @@ class BeakerDriver(PoolDriver):
         self,
         logger: gluetool.log.ContextAdapter
     ) -> Result[PoolResourcesMetrics, Failure]:
-        r_resources = super(BeakerDriver, self).fetch_pool_resources_metrics(logger)
+        r_resources = super().fetch_pool_resources_metrics(logger)
 
         if r_resources.is_error:
             return Error(r_resources.unwrap_error())
@@ -851,7 +1192,10 @@ class BeakerDriver(PoolDriver):
                 raw_machines = ''
 
             else:
-                return Error(failure)
+                return Error(Failure.from_failure(
+                    'failed to fetch system list',
+                    r_query_instances.unwrap_error()
+                ))
 
         else:
             raw_machines = r_query_instances.unwrap().stdout
@@ -868,3 +1212,138 @@ class BeakerDriver(PoolDriver):
         # now.
 
         return Ok(resources)
+
+    def _fetch_avoid_group_hostnames(self, logger: ContextAdapter, groupname: str) -> Result[List[str], Failure]:
+        r_list = self._run_bkr(
+            logger,
+            [
+                'system-list',
+                '--xml-filter', f'<and><group op="=" value="{groupname}"/></and>'
+            ],
+            commandname='bkr.system-list-owned-by-group'
+        )
+
+        if r_list.is_error:
+            return Error(Failure.from_failure(
+                'failed to fetch systems owned by a group',
+                r_list.unwrap_error().update(groupname=groupname)
+            ))
+
+        return Ok([
+            hostname.strip() for hostname in r_list.unwrap().stdout.splitlines()
+        ])
+
+    def refresh_avoid_groups_hostnames(self, logger: ContextAdapter) -> Result[None, Failure]:
+        groups: List[AvoidGroupHostnames] = []
+
+        r_avoid_groups = self.avoid_groups
+
+        if r_avoid_groups.is_error:
+            return Error(r_avoid_groups.unwrap_error())
+
+        for groupname in r_avoid_groups.unwrap():
+            r_list = self._fetch_avoid_group_hostnames(logger, groupname)
+
+            if r_list.is_error:
+                return Error(r_list.unwrap_error())
+
+            groups.append(AvoidGroupHostnames(
+                groupname=groupname,
+                hostnames=r_list.unwrap()
+            ))
+
+        r_refresh = refresh_cached_set(
+            CACHE.get(),
+            self.avoid_groups_hostnames_cache_key,
+            {
+                h.groupname: h
+                for h in groups
+            }
+        )
+
+        if r_refresh.is_error:
+            return Error(r_refresh.unwrap_error())
+
+        return Ok(None)
+
+    def get_avoid_groups_hostnames(self) -> Result[Dict[str, AvoidGroupHostnames], Failure]:
+        return get_cached_set(CACHE.get(), self.avoid_groups_hostnames_cache_key, AvoidGroupHostnames)
+
+    def _get_beaker_machine_log_url(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
+        beaker_logname: str
+    ) -> Result[Optional[str], Failure]:
+        """
+        Extract location (URL) of Beaker machine log.
+
+        :param logger: logger to use for logging.
+        :param guest_request: a request whose logs to look for.
+        :param beaker_logname: a name of the log as known to Beaker (e.g. ``console.log``).
+        :returns: log URL, ``None`` when no such log was found, or a :py:class:`Failure` describing an error.
+        """
+
+        r_job_results = self._get_job_results(logger, BeakerPoolData.unserialize(guest_request).job_id)
+
+        if r_job_results.is_error:
+            return Error(r_job_results.unwrap_error())
+
+        logs = r_job_results.unwrap().select(f'recipe > logs > log[name="{beaker_logname}"]')
+
+        if not logs:
+            return Ok(None)
+
+        return Ok(logs[0]['href'])
+
+    def _update_guest_log_url(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
+        guest_log: GuestLog,
+        beaker_logname: str
+    ) -> Result[GuestLogUpdateProgress, Failure]:
+        r_url = self._get_beaker_machine_log_url(logger, guest_request, beaker_logname)
+
+        if r_url.is_error:
+            return Error(r_url.unwrap_error())
+
+        url = r_url.unwrap()
+
+        if url is None:
+            return Ok(GuestLogUpdateProgress(state=GuestLogState.PENDING))
+
+        return Ok(GuestLogUpdateProgress(state=GuestLogState.COMPLETE, url=url))
+
+    # Since Beaker provides named logs, unlike other drivers, there are updaters for various log names.
+    # To satisfy the "standard" logging expectations, use some of those for common logs like console/url
+    # while keeping their Beaker-native names available at the same time.
+    @guest_log_updater('beaker', 'console', GuestLogContentType.URL)  # type: ignore[arg-type]
+    def _update_guest_log_console_url(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
+        guest_log: GuestLog
+    ) -> Result[GuestLogUpdateProgress, Failure]:
+        return self._update_guest_log_url(logger, guest_request, guest_log, 'console.log')
+
+    @guest_log_updater('beaker', 'console.log', GuestLogContentType.URL)  # type: ignore[arg-type]
+    def _update_guest_log_console_log_url(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
+        guest_log: GuestLog
+    ) -> Result[GuestLogUpdateProgress, Failure]:
+        return self._update_guest_log_url(logger, guest_request, guest_log, 'console.log')
+
+    @guest_log_updater('beaker', 'sys.log', GuestLogContentType.URL)  # type: ignore[arg-type]
+    def _update_guest_log_sys_log_url(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
+        guest_log: GuestLog
+    ) -> Result[GuestLogUpdateProgress, Failure]:
+        return self._update_guest_log_url(logger, guest_request, guest_log, 'sys.log')
+
+
+PoolDriver._drivers_registry['beaker'] = BeakerDriver
