@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import threading
 from typing import Optional
 
@@ -16,6 +19,8 @@ class LocalhostDriver(PoolDriver):
     A dummy driver always "provisioning" a localhost for the given guest.
     """
 
+    drivername = 'localhost'
+
     def acquire_guest(
         self,
         logger: gluetool.log.ContextAdapter,
@@ -42,3 +47,6 @@ class LocalhostDriver(PoolDriver):
         guest_request: GuestRequest
     ) -> Result[bool, Failure]:
         return Ok(True)
+
+
+PoolDriver._drivers_registry['localhost'] = LocalhostDriver
