@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import dataclasses
 import datetime
 import re
@@ -10,14 +13,16 @@ import jq
 import sqlalchemy.orm.session
 from gluetool.result import Error, Ok, Result
 
-from .. import Failure, JSONType, log_dict_yaml
+from .. import Failure, JSONType, log_dict_yaml, process_output_to_str
 from ..db import GuestLog, GuestLogContentType, GuestLogState, GuestRequest, SnapshotRequest
-from ..environment import UNITS, Environment, Flavor, FlavorCpu, FlavorDisk, FlavorDisks
+from ..environment import UNITS, Environment, Flavor, FlavorBoot, FlavorCpu, FlavorDisk, FlavorDisks, \
+    FlavorVirtualization
 from ..knobs import Knob
 from ..metrics import PoolMetrics, PoolNetworkResources, PoolResourcesMetrics, ResourceType
-from . import KNOB_UPDATE_GUEST_REQUEST_TICK, ConsoleUrlData, GuestLogUpdateProgress, HookImageInfoMapper, \
-    PoolCapabilities, PoolData, PoolDriver, PoolImageInfo, PoolImageSSHInfo, PoolResourcesIDs, ProvisioningProgress, \
-    ProvisioningState, SerializedPoolResourcesIDs, create_tempfile, run_cli_tool, test_cli_error
+from . import KNOB_UPDATE_GUEST_REQUEST_TICK, CLIErrorCauses, ConsoleUrlData, GuestLogUpdateProgress, \
+    HookImageInfoMapper, PoolCapabilities, PoolData, PoolDriver, PoolImageInfo, PoolImageSSHInfo, PoolResourcesIDs, \
+    ProvisioningProgress, ProvisioningState, SerializedPoolResourcesIDs, create_tempfile, guest_log_updater, \
+    run_cli_tool
 
 KNOB_BUILD_TIMEOUT: Knob[int] = Knob(
     'openstack.build-timeout',
@@ -47,8 +52,57 @@ KNOB_CONSOLE_BLOB_UPDATE_TICK: Knob[int] = Knob(
     default=30
 )
 
-MISSING_INSTANCE_ERROR_PATTERN = re.compile(r'^No server with a name or ID')
-INSTANCE_NOT_READY_ERROR_PATTERN = re.compile(r'^Instance [a-z0-9\-]+ is not ready')
+KNOB_ENVIRONMENT_TO_IMAGE_MAPPING_FILEPATH: Knob[str] = Knob(
+    'openstack.mapping.environment-to-image.pattern-map.filepath',
+    'Path to a pattern map file with environment to image mapping.',
+    has_db=False,
+    per_pool=True,
+    envvar='ARTEMIS_OPENSTACK_ENVIRONMENT_TO_IMAGE_MAPPING_FILEPATH',
+    cast_from_str=str,
+    default='artemis-image-map-openstack.yaml'
+)
+
+KNOB_ENVIRONMENT_TO_IMAGE_MAPPING_NEEDLE: Knob[str] = Knob(
+    'openstack.mapping.environment-to-image.pattern-map.needle',
+    'A pattern for needle to match in environment to image mapping file.',
+    has_db=False,
+    per_pool=True,
+    envvar='ARTEMIS_OPENSTACK_ENVIRONMENT_TO_IMAGE_MAPPING_NEEDLE',
+    cast_from_str=str,
+    default='{{ os.compose }}'
+)
+
+
+class OsErrorCauses(CLIErrorCauses):
+    NONE = 'none'
+    NO_SUCH_COMMAND = 'no-such-command'
+    MISSING_INSTANCE = 'missing-instance'
+    INSTANCE_NOT_READY = 'instance-not-ready'
+
+
+CLI_ERROR_PATTERNS = {
+    OsErrorCauses.NO_SUCH_COMMAND: re.compile(r'openstack: .+ is not an openstack command'),
+    OsErrorCauses.MISSING_INSTANCE: re.compile(r'^No server with a name or ID'),
+    OsErrorCauses.INSTANCE_NOT_READY: re.compile(r'^Instance [a-z0-9\-]+ is not ready')
+}
+
+
+def os_error_cause_extractor(output: gluetool.utils.ProcessOutput) -> OsErrorCauses:
+    if output.exit_code == 0:
+        return OsErrorCauses.NONE
+
+    stdout = process_output_to_str(output, stream='stdout')
+    stderr = process_output_to_str(output, stream='stderr')
+
+    for cause, pattern in CLI_ERROR_PATTERNS.items():
+        if stdout and pattern.match(stdout):
+            return cause
+
+        if stderr and pattern.match(stderr):
+            return cause
+
+    return OsErrorCauses.NONE
+
 
 # IP address is suplied in a list of mappings
 JQ_QUERY_INSTANCE_IPV4_ADDRESS = jq.compile('.addresses | to_entries[0].value[0]')
@@ -65,13 +119,17 @@ class OpenStackPoolResourcesIDs(PoolResourcesIDs):
 
 
 class OpenStackDriver(PoolDriver):
+    drivername = 'openstack'
+
+    pool_data_class = OpenStackPoolData
+
     def __init__(
         self,
         logger: gluetool.log.ContextAdapter,
         poolname: str,
         pool_config: Dict[str, Any]
     ) -> None:
-        super(OpenStackDriver, self).__init__(logger, poolname, pool_config)
+        super().__init__(logger, poolname, pool_config)
 
         self._os_cmd_base = [
             'openstack',
@@ -137,7 +195,8 @@ class OpenStackDriver(PoolDriver):
             json_output=json_format,
             command_scrubber=lambda cmd: (['openstack'] + options),
             poolname=self.poolname,
-            commandname=commandname
+            commandname=commandname,
+            cause_extractor=os_error_cause_extractor
         )
 
         if r_run.is_error:
@@ -145,10 +204,11 @@ class OpenStackDriver(PoolDriver):
 
             # Detect "instance does not exist" - this error is clearly irrecoverable. No matter how often we would
             # run this method, we would never evenr made it remove instance that doesn't exist.
-            if test_cli_error(failure, MISSING_INSTANCE_ERROR_PATTERN):
+            if failure.command_output \
+               and os_error_cause_extractor(failure.command_output) == OsErrorCauses.MISSING_INSTANCE:
                 failure.recoverable = False
 
-                PoolMetrics.inc_error(self.poolname, 'missing-instance')
+                PoolMetrics.inc_error(self.poolname, OsErrorCauses.MISSING_INSTANCE.value)
 
             return Error(failure)
 
@@ -176,12 +236,12 @@ class OpenStackDriver(PoolDriver):
     ) -> Result[PoolImageInfo, Failure]:
         return self._map_image_name_to_image_info_by_cache(logger, imagename)
 
-    def _env_to_flavor(
+    def _env_to_flavor_or_none(
         self,
         logger: gluetool.log.ContextAdapter,
         session: sqlalchemy.orm.session.Session,
         guest_request: GuestRequest
-    ) -> Result[Flavor, Failure]:
+    ) -> Result[Optional[Flavor], Failure]:
         r_suitable_flavors = self._map_environment_to_flavor_info_by_cache_by_constraints(
             logger,
             guest_request.environment
@@ -193,13 +253,27 @@ class OpenStackDriver(PoolDriver):
         suitable_flavors = r_suitable_flavors.unwrap()
 
         if not suitable_flavors:
+            if self.pool_config.get('use-default-flavor-when-no-suitable', True):
+                guest_request.log_warning_event(
+                    logger,
+                    session,
+                    'no suitable flavors, using default',
+                    poolname=self.poolname
+                )
+
+                return self._map_environment_to_flavor_info_by_cache_by_name_or_none(
+                    logger,
+                    self.pool_config['default-flavor']
+                )
+
             guest_request.log_warning_event(
                 logger,
                 session,
-                'no suitable flavors, using default'
+                'no suitable flavors',
+                poolname=self.poolname
             )
 
-            return self._map_environment_to_flavor_info_by_cache_by_name(logger, self.pool_config['default-flavor'])
+            return Ok(None)
 
         if self.pool_config['default-flavor'] in [flavor.name for flavor in suitable_flavors]:
             logger.info('default flavor among suitable ones, using it')
@@ -212,6 +286,24 @@ class OpenStackDriver(PoolDriver):
 
         return Ok(suitable_flavors[0])
 
+    def _env_to_flavor(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        session: sqlalchemy.orm.session.Session,
+        guest_request: GuestRequest
+    ) -> Result[Flavor, Failure]:
+        r_flavor = self._env_to_flavor_or_none(logger, session, guest_request)
+
+        if r_flavor.is_error:
+            return Error(r_flavor.unwrap_error())
+
+        flavor = r_flavor.unwrap()
+
+        if flavor is None:
+            return Error(Failure('no suitable flavor'))
+
+        return Ok(flavor)
+
     def _env_to_network(self, environment: Environment) -> Result[Any, Failure]:
         metrics = PoolResourcesMetrics(self.poolname)
         metrics.sync()
@@ -258,7 +350,10 @@ class OpenStackDriver(PoolDriver):
         r_output = self._run_os(os_options, commandname='os.server-show')
 
         if r_output.is_error:
-            return Error(r_output.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch instance information',
+                r_output.unwrap_error()
+            ))
 
         return Ok(r_output.unwrap())
 
@@ -271,7 +366,10 @@ class OpenStackDriver(PoolDriver):
         r_output = self._run_os(os_options, commandname='os.image-show')
 
         if r_output.is_error:
-            return Error(r_output.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch snapshot information',
+                r_output.unwrap_error()
+            ))
 
         return Ok(r_output.unwrap())
 
@@ -331,10 +429,10 @@ class OpenStackDriver(PoolDriver):
 
             tags = r_tags.unwrap()
 
-            property_options: List[str] = sum([
+            property_options: List[str] = sum((
                 ['--property', f'{tag}={value}']
                 for tag, value in tags.items()
-            ], [])
+            ), [])
 
             os_options = [
                 'server',
@@ -426,7 +524,10 @@ class OpenStackDriver(PoolDriver):
         r_stop = self._run_os(os_options, json_format=False, commandname='os.server-stop')
 
         if r_stop.is_error:
-            return Error(r_stop.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to stop instance',
+                r_stop.unwrap_error()
+            ))
 
         return Ok(True)
 
@@ -440,6 +541,10 @@ class OpenStackDriver(PoolDriver):
         r_start = self._run_os(os_options, json_format=False, commandname='os.server-start')
 
         if r_start.is_error:
+            return Error(Failure.from_failure(
+                'failed to start instance',
+                r_start.unwrap_error()
+            ))
             return Error(r_start.unwrap_error())
 
         return Ok(True)
@@ -450,7 +555,7 @@ class OpenStackDriver(PoolDriver):
         session: sqlalchemy.orm.session.Session,
         guest_request: GuestRequest
     ) -> Result[bool, Failure]:
-        r_answer = super(OpenStackDriver, self).can_acquire(logger, session, guest_request)
+        r_answer = super().can_acquire(logger, session, guest_request)
 
         if r_answer.is_error:
             return Error(r_answer.unwrap_error())
@@ -470,6 +575,9 @@ class OpenStackDriver(PoolDriver):
         if r_flavor.is_error:
             return Error(r_flavor.unwrap_error())
 
+        if r_flavor.unwrap() is None:
+            return Ok(False)
+
         return Ok(True)
 
     def acquire_console_url(
@@ -487,7 +595,10 @@ class OpenStackDriver(PoolDriver):
         ]
         r_output = self._run_os(os_options, commandname='os.console-url-show')
         if r_output.is_error:
-            return Error(r_output.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch console URL',
+                r_output.unwrap_error()
+            ))
         # NOTE(ivasilev) The following cast is needed to keep quiet the typing check
         data = cast(Dict[str, str], r_output.unwrap())
 
@@ -516,7 +627,10 @@ class OpenStackDriver(PoolDriver):
         r_output = self._run_os(os_options, commandname='os.server-image-create')
 
         if r_output.is_error:
-            return Error(r_output.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to create snapshot',
+                r_output.unwrap_error()
+            ))
 
         return Ok(ProvisioningProgress(
             state=ProvisioningState.PENDING,
@@ -570,7 +684,10 @@ class OpenStackDriver(PoolDriver):
         r_output = self._run_os(os_options, json_format=False, commandname='os.image-delete')
 
         if r_output.is_error:
-            return Error(r_output.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to delete snapshot',
+                r_output.unwrap_error()
+            ))
 
         return Ok(True)
 
@@ -589,7 +706,10 @@ class OpenStackDriver(PoolDriver):
         r_output = self._run_os(os_options, json_format=False, commandname='os.server-rebuild')
 
         if r_output.is_error:
-            return Error(r_output.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to rebuild instance',
+                r_output.unwrap_error()
+            ))
 
         return Ok(True)
 
@@ -729,7 +849,7 @@ class OpenStackDriver(PoolDriver):
         logger: gluetool.log.ContextAdapter,
         raw_resource_ids: SerializedPoolResourcesIDs
     ) -> Result[None, Failure]:
-        resource_ids = OpenStackPoolResourcesIDs.unserialize(raw_resource_ids)
+        resource_ids = OpenStackPoolResourcesIDs.unserialize_from_json(raw_resource_ids)
 
         if resource_ids.instance_id:
             r_output = self._run_os([
@@ -745,7 +865,10 @@ class OpenStackDriver(PoolDriver):
                 failure = r_output.unwrap_error()
                 failure.fail_guest_request = False
 
-                return Error(failure)
+                return Error(Failure.from_failure(
+                    'failed to delete instance',
+                    failure
+                ))
 
             self.inc_costs(logger, ResourceType.VIRTUAL_MACHINE, resource_ids.ctime)
 
@@ -755,7 +878,7 @@ class OpenStackDriver(PoolDriver):
         self,
         logger: gluetool.log.ContextAdapter
     ) -> Result[PoolResourcesMetrics, Failure]:
-        r_resources = super(OpenStackDriver, self).fetch_pool_resources_metrics(logger)
+        r_resources = super().fetch_pool_resources_metrics(logger)
 
         if r_resources.is_error:
             return Error(r_resources.unwrap_error())
@@ -769,7 +892,10 @@ class OpenStackDriver(PoolDriver):
         )
 
         if r_query_limits.is_error:
-            return Error(r_query_limits.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch tenant limits',
+                r_query_limits.unwrap_error()
+            ))
 
         raw_limits_container = r_query_limits.unwrap()
 
@@ -794,6 +920,9 @@ class OpenStackDriver(PoolDriver):
             elif name == 'totalSnapshotsUsed':
                 resources.usage.snapshots = int(value)
 
+            elif name == 'totalGigabytesUsed':
+                resources.usage.diskspace = int(value) * 1073741824
+
             # When updating limits, make sure to not overwrite those already specified by pool configuration.
             elif name == 'maxTotalCores' and resources.limits.cores is None:
                 resources.limits.cores = int(value)
@@ -808,6 +937,9 @@ class OpenStackDriver(PoolDriver):
             elif name == 'maxTotalSnapshots' and resources.limits.snapshots is None:
                 resources.limits.snapshots = int(value)
 
+            elif name == 'maxTotalVolumeGigabytes' and resources.limits.diskspace is None:
+                resources.limits.diskspace = int(value) * 1073741824
+
         r_networks = self._run_os([
             'ip',
             'availability',
@@ -816,7 +948,10 @@ class OpenStackDriver(PoolDriver):
         ], json_format=True, commandname='os.ip-availability-list')
 
         if r_networks.is_error:
-            return Error(r_networks.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch network information',
+                r_networks.unwrap_error()
+            ))
 
         # networks have the following structure:
         # [
@@ -846,13 +981,17 @@ class OpenStackDriver(PoolDriver):
         r_images = self._run_os(['image', 'list'], commandname='os.image-list')
 
         if r_images.is_error:
-            return Error(r_images.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch image information',
+                r_images.unwrap_error()
+            ))
 
         try:
             return Ok([
                 PoolImageInfo(
                     name=image['Name'],
                     id=image['ID'],
+                    boot=FlavorBoot(),
                     ssh=PoolImageSSHInfo()
                 )
                 for image in cast(List[Dict[str, str]], r_images.unwrap())
@@ -883,7 +1022,10 @@ class OpenStackDriver(PoolDriver):
         r_flavors = self._run_os(['flavor', 'list'], commandname='os.flavor-list')
 
         if r_flavors.is_error:
-            return Error(r_flavors.unwrap_error())
+            return Error(Failure.from_failure(
+                'failed to fetch flavor information',
+                r_flavors.unwrap_error()
+            ))
 
         if self.pool_config.get('flavor-regex'):
             flavor_name_pattern: Optional[Pattern[str]] = re.compile(self.pool_config['flavor-regex'])
@@ -900,13 +1042,16 @@ class OpenStackDriver(PoolDriver):
                         cores=int(flavor['VCPUs'])
                     ),
                     # memory is reported in MiB
-                    memory=int(flavor['RAM']) * UNITS.mebibytes,
+                    memory=UNITS.Quantity(int(flavor['RAM']), UNITS.mebibytes),
                     disk=FlavorDisks([
                         FlavorDisk(
                             # diskspace is reported in GiB
-                            size=int(flavor['Disk']) * UNITS.gibibytes
+                            size=UNITS.Quantity(int(flavor['Disk']), UNITS.gibibytes)
                         )
-                    ])
+                    ]),
+                    virtualization=FlavorVirtualization(
+                        is_virtualized=True
+                    )
                 )
                 for flavor in cast(List[Dict[str, str]], r_flavors.unwrap())
                 if flavor_name_pattern is None or flavor_name_pattern.match(flavor['Name'])
@@ -919,17 +1064,36 @@ class OpenStackDriver(PoolDriver):
                 flavor_info=r_flavors.unwrap()
             ))
 
-    def update_guest_log(
+    def _do_fetch_console(
         self,
         guest_request: GuestRequest,
+        resource: str,
+        json_format: bool = True
+    ) -> Result[Optional[JSONType], Failure]:
+        r_output = self._run_os([
+            'console',
+            resource,
+            'show',
+            OpenStackPoolData.unserialize(guest_request).instance_id
+        ], json_format=json_format, commandname=f'console-{resource}-show')
+
+        if r_output.is_error:
+            failure = r_output.unwrap_error()
+
+            # Detect "instance not ready".
+            if failure.command_output \
+               and os_error_cause_extractor(failure.command_output) == OsErrorCauses.INSTANCE_NOT_READY:
+                return Ok(None)
+
+        return r_output
+
+    @guest_log_updater('openstack', 'console', GuestLogContentType.URL)  # type: ignore[arg-type]
+    def _update_guest_log_console_url(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
         guest_log: GuestLog
     ) -> Result[GuestLogUpdateProgress, Failure]:
-        if guest_log.logname != 'console':
-            # anything but "console" is unsupported so far
-            return Ok(GuestLogUpdateProgress(
-                state=GuestLogState.ERROR
-            ))
-
         r_delay_update = KNOB_CONSOLE_BLOB_UPDATE_TICK.get_value(poolname=self.poolname)
 
         if r_delay_update.is_error:
@@ -937,51 +1101,43 @@ class OpenStackDriver(PoolDriver):
 
         delay_update = r_delay_update.unwrap()
 
-        def _do_fetch(resource: str, json_format: bool = True) -> Result[Optional[JSONType], Failure]:
-            r_output = self._run_os([
-                'console',
-                resource,
-                'show',
-                OpenStackPoolData.unserialize(guest_request).instance_id
-            ], json_format=json_format, commandname=f'console-{resource}-show')
-
-            if r_output.is_error:
-                failure = r_output.unwrap_error()
-
-                # Detect "instance not ready".
-                if test_cli_error(failure, INSTANCE_NOT_READY_ERROR_PATTERN):
-                    return Ok(None)
-
-            return r_output
-
-        if guest_log.contenttype == GuestLogContentType.URL:
-            r_output = _do_fetch('url')
-
-            if r_output.is_error:
-                return Error(r_output.unwrap_error())
-
-            output = r_output.unwrap()
-
-            if output is None:
-                return Ok(GuestLogUpdateProgress(
-                    state=GuestLogState.IN_PROGRESS,
-                    delay_update=delay_update
-                ))
-
-            return Ok(GuestLogUpdateProgress(
-                state=GuestLogState.COMPLETE,
-                url=cast(Dict[str, str], output)['url'],
-                expires=datetime.datetime.utcnow() + datetime.timedelta(seconds=KNOB_CONSOLE_URL_EXPIRES.value)
-            ))
-
-        # We're left with console/blob.
-        r_output = _do_fetch('log', json_format=False)
+        r_output = self._do_fetch_console(guest_request, 'url')
 
         if r_output.is_error:
             return Error(r_output.unwrap_error())
 
-        # TODO logs: do some sort of magic to find out whether the blob we just got is already in DB.
-        # Maybe use difflib, or use delimiters and timestamps. We do not want to overide what we already have.
+        output = r_output.unwrap()
+
+        if output is None:
+            return Ok(GuestLogUpdateProgress(
+                state=GuestLogState.IN_PROGRESS,
+                delay_update=delay_update
+            ))
+
+        return Ok(GuestLogUpdateProgress(
+            state=GuestLogState.COMPLETE,
+            url=cast(Dict[str, str], output)['url'],
+            expires=datetime.datetime.utcnow() + datetime.timedelta(seconds=KNOB_CONSOLE_URL_EXPIRES.value)
+        ))
+
+    @guest_log_updater('openstack', 'console', GuestLogContentType.BLOB)  # type: ignore[arg-type]
+    def _update_guest_log_console_blob(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        guest_request: GuestRequest,
+        guest_log: GuestLog
+    ) -> Result[GuestLogUpdateProgress, Failure]:
+        r_delay_update = KNOB_CONSOLE_BLOB_UPDATE_TICK.get_value(poolname=self.poolname)
+
+        if r_delay_update.is_error:
+            return Error(r_delay_update.unwrap_error())
+
+        delay_update = r_delay_update.unwrap()
+
+        r_output = self._do_fetch_console(guest_request, 'log', json_format=False)
+
+        if r_output.is_error:
+            return Error(r_output.unwrap_error())
 
         output = r_output.unwrap()
 
@@ -997,3 +1153,6 @@ class OpenStackDriver(PoolDriver):
             blob=cast(str, output),
             delay_update=delay_update
         ))
+
+
+PoolDriver._drivers_registry['openstack'] = OpenStackDriver
