@@ -1,8 +1,12 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 """
 Guest request state definitions and helpers.
 """
 
 import enum
+from typing import cast
 
 import gluetool.log
 
@@ -71,10 +75,20 @@ class GuestLogger(gluetool.log.ContextAdapter):
         :param guestname: snapshot request name to add.
         """
 
-        super(GuestLogger, self).__init__(logger, {
+        super().__init__(logger, {
             'ctx_guest_name': (10, guestname)
         })
 
+    @property
+    def guestname(self) -> str:
+        """
+        Return guest request name tracked by the logger.
+
+        :returns: name of the guest request to propagate into logging context.
+        """
+
+        return cast(str, self._contexts['guest_name'][1])
+
 
 class SnapshotLogger(gluetool.log.ContextAdapter):
     """
@@ -89,6 +103,16 @@ class SnapshotLogger(gluetool.log.ContextAdapter):
         :param snapshotname: snapshot request name to add.
         """
 
-        super(SnapshotLogger, self).__init__(logger, {
+        super().__init__(logger, {
             'ctx_snapshot_name': (11, snapshotname)
         })
+
+    @property
+    def snapshotname(self) -> str:
+        """
+        Return snapshot name tracked by the logger.
+
+        :returns: name of the snapshot to propagate into logging context.
+        """
+
+        return cast(str, self._contexts['snapshot_name'][1])
