@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import inspect
 import logging
 import os
@@ -55,7 +58,7 @@ class KnobSourceEnv(KnobSource[T]):
     """
 
     def __init__(self, knob: 'Knob[T]', envvar: str) -> None:
-        super(KnobSourceEnv, self).__init__(knob)
+        super().__init__(knob)
 
         self.envvar = envvar
 
@@ -141,7 +144,7 @@ class KnobSourceDefault(KnobSource[T]):
     """
 
     def __init__(self, knob: 'Knob[T]', default: T, default_label: Optional[str] = None) -> None:
-        super(KnobSourceDefault, self).__init__(knob)
+        super().__init__(knob)
 
         self.default = default
         self.default_label = default_label
@@ -202,7 +205,7 @@ class KnobSourceDBGlobal(KnobSourceDB[T]):
     Read knob value from a database.
     """
 
-    def get_value(  # type: ignore  # match parent
+    def get_value(  # type: ignore[override]  # match parent
         self,
         *,
         session: Session,
@@ -222,7 +225,7 @@ class KnobSourceDBPerPool(KnobSourceDB[T]):
     * ``${original knob name}``
     """
 
-    def get_value(  # type: ignore  # match parent
+    def get_value(  # type: ignore[override]  # match parent
         self,
         *,
         session: Session,
@@ -258,7 +261,7 @@ class KnobSourceActual(KnobSource[T]):
     """
 
     def __init__(self, knob: 'Knob[T]', value: T) -> None:
-        super(KnobSourceActual, self).__init__(knob)
+        super().__init__(knob)
 
         self.value = value
 
@@ -273,7 +276,7 @@ class KnobSourceActual(KnobSource[T]):
 
 class KnobError(ValueError):
     def __init__(self, knob: 'Knob[T]', message: str, failure: Optional['Failure'] = None) -> None:
-        super(KnobError, self).__init__(f'Badly configured knob: {message}')
+        super().__init__(f'Badly configured knob: {message}')
 
         self.knobname = knob.knobname
         self.failure = failure
@@ -478,7 +481,7 @@ class Knob(Generic[T]):
         if self.cast_from_str:
             traits += [f'cast-from-str={self.cast_from_str.__name__}']
 
-        traits += sum([source.to_repr() for source in self._sources], [])
+        traits += sum((source.to_repr() for source in self._sources), [])
 
         return f'<Knob: {self.knobname}: {" ".join(traits)}>'
 
@@ -558,7 +561,7 @@ class Knob(Generic[T]):
         """
         For a given knobname - which belongs to a knob with per-pool capability - find its "parent" knob.
 
-        Per-pool knobs don't have 1:1 mapping between a Python :py:ref:`Knob` instance and its DB record.
+        Per-pool knobs don't have 1:1 mapping between a Python :py:class:`Knob` instance and its DB record.
         But the "parent" knob, the one actually declared somewhere in the source, can be found by name
         after stripping the pool name from the given knob name.
         """
@@ -611,7 +614,10 @@ KNOB_CONFIG_DIRPATH: Knob[str] = Knob(
 
 KNOB_BROKER_URL: Knob[str] = Knob(
     'broker.url',
-    'Broker URL.',
+    """
+    Broker URL. See https://pika.readthedocs.io/en/1.2.0/modules/parameters.html#pika.connection.URLParameters
+    for full list of connection parameters that can be specified via URL.
+    """,
     has_db=False,
     envvar='ARTEMIS_BROKER_URL',
     cast_from_str=str,
@@ -638,18 +644,6 @@ KNOB_BROKER_CONFIRM_DELIVERY: Knob[bool] = Knob(
     default=True
 )
 
-KNOB_BROKER_HEARTBEAT_TIMEOUT: Knob[int] = Knob(
-    'broker.heartbeat-timeout',
-    """
-    An interval, in seconds, after which a broker client should ping the server over the established connection to
-    keep both parties aware the connection should be kept alive.
-    """,
-    has_db=False,
-    envvar='ARTEMIS_BROKER_HEARTBEAT_TIMEOUT',
-    cast_from_str=int,
-    default=60
-)
-
 KNOB_DB_URL: Knob[str] = Knob(
     'db.url',
     'Database URL.',
@@ -674,7 +668,8 @@ KNOB_VAULT_PASSWORD_FILEPATH: Knob[str] = Knob(
     has_db=False,
     envvar='ARTEMIS_VAULT_PASSWORD_FILE',
     cast_from_str=lambda s: os.path.expanduser(s.strip()),
-    default=os.path.expanduser('~/.vault_password')
+    default=os.path.expanduser('~/.vault_password'),
+    default_label='$HOME/.vault_password'
 )
 
 KNOB_LOGGING_DB_QUERIES: Knob[bool] = Knob(
@@ -747,6 +742,122 @@ KNOB_POOL_ENABLED: Knob[bool] = Knob(
     default=True
 )
 
+KNOB_WORKER_PROCESS_METRICS_ENABLED: Knob[bool] = Knob(
+    'worker.metrics.process.enabled',
+    'If enabled, various metrics related to worker processes would be collected.',
+    has_db=False,
+    per_pool=False,
+    envvar='ARTEMIS_WORKER_PROCESS_METRICS_ENABLED',
+    cast_from_str=gluetool.utils.normalize_bool_option,
+    default=True
+)
+
+KNOB_WORKER_PROCESS_METRICS_UPDATE_TICK: Knob[int] = Knob(
+    'worker.metrics.process.update-tick',
+    'How often, in seconds, should workers update their process metrics cache.',
+    has_db=False,
+    per_pool=False,
+    envvar='ARTEMIS_WORKER_PROCESS_METRICS_UPDATE_TICK',
+    cast_from_str=int,
+    default=60
+)
+
+KNOB_WORKER_PROCESS_METRICS_TTL: Knob[int] = Knob(
+    'worker.metrics.process.ttl',
+    'How long, in seconds, should worker process metrics remain in cache.',
+    has_db=False,
+    per_pool=False,
+    envvar='ARTEMIS_WORKER_PROCESS_METRICS_TTL',
+    cast_from_str=int,
+    default=120
+)
+
+KNOB_WORKER_TRAFFIC_METRICS_ENABLED: Knob[bool] = Knob(
+    'worker.metrics.traffic.enabled',
+    'If enabled, various metrics related to tasks and requests would be collected.',
+    has_db=False,
+    per_pool=False,
+    envvar='ARTEMIS_WORKER_TRAFFIC_METRICS_ENABLED',
+    cast_from_str=gluetool.utils.normalize_bool_option,
+    default=True
+)
+
+KNOB_WORKER_TRAFFIC_METRICS_TTL: Knob[int] = Knob(
+    'worker.metrics.traffic.ttl',
+    'How long, in seconds, should worker traffic metrics remain in cache.',
+    has_db=False,
+    per_pool=False,
+    envvar='ARTEMIS_WORKER_TRAFFIC_METRICS_TTL',
+    cast_from_str=int,
+    # The value should be comparable to how long tasks can take, which depends on resources available to workers.
+    default=600
+)
+
+KNOB_DEPLOYMENT: Knob[str] = Knob(
+    'deployment.name',
+    'Optional name of the Artemis deployment (e.g. "production-01" or "development").',
+    has_db=False,
+    envvar='ARTEMIS_DEPLOYMENT',
+    cast_from_str=str,
+    default='undefined-deployment'
+)
+
+KNOB_COMPONENT: Knob[str] = Knob(
+    'deployment.component',
+    'Optional name of the Artemis component (e.g. "worker", "api", etc.).',
+    has_db=False,
+    envvar='ARTEMIS_COMPONENT',
+    cast_from_str=str,
+    default='undefined-component'
+)
+
+KNOB_DEPLOYMENT_ENVIRONMENT: Knob[str] = Knob(
+    'deployment.environment',
+    'Optional environment of the Artemis deployment (e.g. "production" or "staging").',
+    has_db=False,
+    envvar='ARTEMIS_DEPLOYMENT_ENVIRONMENT',
+    cast_from_str=str,
+    default='undefined-deployment-environment'
+)
+
+KNOB_SENTRY_DSN: Knob[Optional[str]] = Knob(
+    'sentry.dsn',
+    'Sentry DSN.',
+    has_db=False,
+    envvar='ARTEMIS_SENTRY_DSN',
+    cast_from_str=str,
+    # TODO: Knob cannot use None as actual default value. Needs a fix.
+    default='undefined'
+)
+
+KNOB_SENTRY_BASE_URL: Knob[Optional[str]] = Knob(
+    'sentry.base-url',
+    'Sentry base URL, for nice event URLs in logs.',
+    has_db=False,
+    envvar='ARTEMIS_SENTRY_BASE_URL',
+    cast_from_str=str,
+    # TODO: Knob cannot use None as actual default value. Needs a fix.
+    default='undefined'
+)
+
+KNOB_SENTRY_DISABLE_CERT_VERIFICATION: Knob[bool] = Knob(
+    'sentry.disable-cert-verification',
+    'When enabled, Artemis would disable HTTPS certificate verification when submitting to Sentry.',
+    has_db=False,
+    envvar='ARTEMIS_SENTRY_DISABLE_CERT_VERIFICATION',
+    cast_from_str=gluetool.utils.normalize_bool_option,
+    default=False
+)
+
+KNOB_LOGGING_SENTRY: Knob[bool] = Knob(
+    'logging.sentry',
+    'When enabled, Artemis would log more Sentry-related debug info.',
+    has_db=False,
+    envvar='ARTEMIS_LOG_SENTRY',
+    cast_from_str=gluetool.utils.normalize_bool_option,
+    default=False
+)
+
 
 def get_vault_password() -> str:
     password = KNOB_VAULT_PASSWORD.value
@@ -754,5 +865,5 @@ def get_vault_password() -> str:
     if password:
         return password
 
-    with open(KNOB_VAULT_PASSWORD_FILEPATH.value, 'r') as f:
+    with open(KNOB_VAULT_PASSWORD_FILEPATH.value) as f:
         return f.read()
