@@ -1,20 +1,27 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import datetime
-import inspect
+import os
+import threading
 import traceback
-from typing import TYPE_CHECKING, Any, Dict, Optional, Set, Union, cast
+from typing import TYPE_CHECKING, Any, Dict, Optional, Set, cast
 
 import dramatiq.broker
 import dramatiq.message
 import dramatiq.middleware
 import dramatiq.middleware.retries
+import dramatiq.worker
 import gluetool.log
+import redis
 from dramatiq.common import compute_backoff, current_millis
+from gluetool.result import Ok
 
 from .guest import GuestLogger
 
 if TYPE_CHECKING:
     from . import ExceptionInfoType
-    from .tasks import Actor
+    from .tasks import Actor, ActorArgumentType, TaskCall
 
 
 # Dramatiq does not have a global default for maximal number of retries, the value is only present as a default
@@ -22,74 +29,57 @@ if TYPE_CHECKING:
 DEFAULT_MAX_RETRIES = 20
 
 
-def _actor_arguments(
-    logger: gluetool.log.ContextAdapter,
-    message: dramatiq.message.Message,
-    actor: 'Actor'
-) -> Dict[str, Union[str, None]]:
-    signature = inspect.signature(actor.fn)
+def _dump_message(message: dramatiq.message.Message) -> Dict[str, Any]:
+    # message.asdict() is nice, but returns ordered dict. we don't care about
+    # order, and OrderedDict would need extra care when serializing anywhere,
+    # so convert it to plain dict.
 
-    gluetool.log.log_dict(logger.debug, 'raw message data', message._message)
-
-    if len(signature.parameters) != len(message._message[2]):
-        from . import Failure
-
-        Failure(
-            'actor signature parameters does not match message content',
-            signature=[name for name in signature.parameters.keys()],
-            arguments=[repr(arg) for arg in message._message[2]]
-        ).handle(logger)
-
-        return {}
-
-    return {
-        name: message._message[2][index]
-        for index, name in enumerate(signature.parameters.keys())
-    }
+    return dict(**message.asdict())
 
 
 def _get_message_limit(
     message: dramatiq.broker.MessageProxy,
-    actor: 'Actor',
     key: str,
-    default: int
+    default: int,
+    actor: Optional['Actor'] = None,
 ) -> int:
     value = cast(Optional[int], message.options.get(key))
 
     if value:
         return value
 
-    value = cast(Optional[int], actor.options.get(key))
+    if actor:
+        value = cast(Optional[int], actor.options.get(key))
 
-    if value:
-        return value
+        if value:
+            return value
 
     return default
 
 
-def _message_max_retries(message: dramatiq.broker.MessageProxy, actor: 'Actor') -> int:
-    return _get_message_limit(message, actor, 'max_retries', DEFAULT_MAX_RETRIES)
+def _message_max_retries(message: dramatiq.broker.MessageProxy, actor: Optional['Actor'] = None) -> int:
+    return _get_message_limit(message, 'max_retries', DEFAULT_MAX_RETRIES, actor=actor)
 
 
-def _message_min_backoff(message: dramatiq.broker.MessageProxy, actor: 'Actor') -> int:
-    return _get_message_limit(message, actor, 'min_backoff', dramatiq.middleware.retries.DEFAULT_MIN_BACKOFF)
+def _message_min_backoff(message: dramatiq.broker.MessageProxy, actor: Optional['Actor'] = None) -> int:
+    return _get_message_limit(message, 'min_backoff', dramatiq.middleware.retries.DEFAULT_MIN_BACKOFF, actor=actor)
 
 
-def _message_max_backoff(message: dramatiq.broker.MessageProxy, actor: 'Actor') -> int:
-    return _get_message_limit(message, actor, 'max_backoff', dramatiq.middleware.retries.DEFAULT_MAX_BACKOFF)
+def _message_max_backoff(message: dramatiq.broker.MessageProxy, actor: Optional['Actor'] = None) -> int:
+    return _get_message_limit(message, 'max_backoff', dramatiq.middleware.retries.DEFAULT_MAX_BACKOFF, actor=actor)
 
 
 def _message_backoff(
     message: dramatiq.broker.MessageProxy,
-    actor: 'Actor',
-    retries: int
+    retries: int,
+    actor: Optional['Actor'] = None
 ) -> int:
     return cast(
         int,
         compute_backoff(
             retries,
-            factor=_message_min_backoff(message, actor),
-            max_backoff=_message_max_backoff(message, actor)
+            factor=_message_min_backoff(message, actor=actor),
+            max_backoff=_message_max_backoff(message, actor=actor)
         )[1]
     )
 
@@ -98,7 +88,7 @@ def _retry_message(
     logger: gluetool.log.ContextAdapter,
     broker: dramatiq.broker.Broker,
     message: dramatiq.message.Message,
-    actor: 'Actor',
+    task_call: Optional['TaskCall'] = None,
     exc_info: Optional['ExceptionInfoType'] = None
 ) -> None:
     """
@@ -111,7 +101,7 @@ def _retry_message(
         message.options['traceback'] = '\n'.join(traceback.format_exception(*exc_info, limit=30))
 
     retries = cast(int, message.options['retries'])
-    backoff = _message_backoff(message, actor, retries)
+    backoff = _message_backoff(message, retries, actor=task_call.actor if task_call else None)
 
     retry_at = datetime.datetime.utcnow() + datetime.timedelta(milliseconds=backoff)
 
@@ -132,6 +122,8 @@ def _fail_message(
 
     from . import Failure
 
+    details['broker_message'] = _dump_message(message)
+
     Failure(error_message, **details).handle(logger)
 
     message.fail()
@@ -140,8 +132,7 @@ def _fail_message(
 def _handle_tails(
     logger: gluetool.log.ContextAdapter,
     message: dramatiq.message.Message,
-    actor: 'Actor',
-    actor_arguments: Dict[str, Optional[str]]
+    task_call: 'TaskCall'
 ) -> bool:
     """
     Handle the "tails": when we run out of retries on a task, we cannot just let it fail, but we must take
@@ -151,15 +142,17 @@ def _handle_tails(
     The task here is to dispatch the correct handler.
     """
 
-    from .tasks import TailHandler, get_root_db
+    from .tasks import get_root_db
 
-    tail_handler = cast(TailHandler, actor.options['tail_handler'])
-    tail_logger = tail_handler.get_logger(logger, actor, actor_arguments)
+    tail_handler = task_call.tail_handler
+    assert tail_handler is not None
+
+    tail_logger = tail_handler.get_logger(logger, task_call)
 
     db = get_root_db(tail_logger)
 
     with db.get_session() as session:
-        if tail_handler.handle_tail(tail_logger, db, session, actor, actor_arguments):
+        if tail_handler.handle_tail(tail_logger, db, session, task_call):
             return True
 
     tail_logger.error('failed to handle the chain tail')
@@ -174,7 +167,7 @@ def _handle_tails(
     return False
 
 
-class Retries(dramatiq.middleware.retries.Retries):  # type: ignore  # Class cannot subclass 'Retries
+class Retries(dramatiq.middleware.retries.Retries):  # type: ignore[misc]  # cannot subclass 'Retries'
     @property
     def actor_options(self) -> Set[str]:
         return {
@@ -202,43 +195,41 @@ class Retries(dramatiq.middleware.retries.Retries):  # type: ignore  # Class can
             return
 
         from . import get_logger
+        from .tasks import TaskCall
 
         logger = get_logger()
 
-        actor = cast('Actor', broker.get_actor(message.actor_name))
+        task_call = TaskCall.from_message(broker, message)
 
         # `retries` key is initialized to 0 - while other fields are optional, this one is expected to exist.
         retries = message.options.setdefault('retries', 0)
-        max_retries = _message_max_retries(message, actor)
-        retry_when = actor.options.get('retry_when', self.retry_when)
+        max_retries = _message_max_retries(message, task_call.actor)
+        retry_when = task_call.actor.options.get('retry_when', self.retry_when)
 
-        actor_arguments = _actor_arguments(logger, message, actor)
+        guestname = task_call.named_args.get('guestname', None)
 
-        guestname = actor_arguments['guestname'] if actor_arguments and 'guestname' in actor_arguments else None
-
-        if guestname:
+        if isinstance(guestname, str):
             logger = GuestLogger(logger, guestname)
 
-        logger.info(f'retries: message={message.message_id} actor={actor.actor_name} current-retries={retries} max-retries={max_retries}')  # noqa
+        logger.info(f'retries: message={message.message_id} actor={task_call.actor.actor_name} current-retries={retries} max-retries={max_retries}')  # noqa: E501
 
         if retry_when is not None and not retry_when(retries, exception) or \
            retry_when is None and max_retries is not None and retries >= max_retries:
 
             # Kill messages for tasks we don't handle in any better way. After all, they did run out of retires.
-            if actor.options.get('tail_handler') is None:
+            if not task_call.has_tail_handler:
                 return _fail_message(
                     logger,
                     message,
-                    f'retries exceeded for message {message.message_id}',
-                    task_name=actor.actor_name,
-                    task_args=actor_arguments,
+                    'retries exceeded',
+                    task_call=task_call,
                     guestname=guestname
                 )
 
-            if _handle_tails(logger, message, actor, actor_arguments) is True:
+            if _handle_tails(logger, message, task_call) is True:
                 return
 
-        _retry_message(logger, broker, message, actor)
+        _retry_message(logger, broker, message, task_call=task_call)
 
 
 MESSAGE_NOTE_OPTION_KEY = 'artemis_notes'
@@ -257,7 +248,10 @@ def set_message_note(note: str, value: str) -> None:
 
     from .context import CURRENT_MESSAGE
 
-    options = cast(Dict[str, Dict[str, str]], CURRENT_MESSAGE.get().options)
+    message_proxy = CURRENT_MESSAGE.get()
+    assert message_proxy is not None
+
+    options = cast(Dict[str, Dict[str, str]], message_proxy.options)
     options.setdefault(MESSAGE_NOTE_OPTION_KEY, {})
     options[MESSAGE_NOTE_OPTION_KEY][note] = value
 
@@ -269,14 +263,17 @@ def get_metric_note(note: str) -> Optional[str]:
 
     from .context import CURRENT_MESSAGE
 
-    options = cast(Dict[str, Dict[str, str]], CURRENT_MESSAGE.get().options)
+    message_proxy = CURRENT_MESSAGE.get()
+    assert message_proxy is not None
+
+    options = cast(Dict[str, Dict[str, str]], message_proxy.options)
 
     return options.get(MESSAGE_NOTE_OPTION_KEY, {}).get(note)
 
 
-class Prometheus(dramatiq.middleware.Middleware):  # type: ignore  # Class cannot subclass 'Middleware'
+class Prometheus(dramatiq.middleware.Middleware):  # type: ignore[misc]  # cannot subclass 'Middleware'
     def __init__(self) -> None:
-        super(Prometheus, self).__init__()
+        super().__init__()
 
         self._delayed_messages: Set[str] = set()
         self._message_start_times: Dict[str, int] = {}
@@ -326,15 +323,12 @@ class Prometheus(dramatiq.middleware.Middleware):  # type: ignore  # Class canno
         result: None = None,
         exception: Optional[BaseException] = None
     ) -> None:
-        from . import get_logger
         from .metrics import TaskMetrics
-
-        logger = get_logger()
+        from .tasks import TaskCall
 
         labels = (message.queue_name, message.actor_name)
-        actor = broker.get_actor(message.actor_name)
 
-        actor_arguments = _actor_arguments(logger, message, actor)
+        task_call = TaskCall.from_message(broker, message)
 
         message_start_time = self._message_start_times.pop(message.message_id, current_millis())
         message_duration = current_millis() - message_start_time
@@ -342,10 +336,10 @@ class Prometheus(dramatiq.middleware.Middleware):  # type: ignore  # Class canno
         # Extract the poolname. `None` is a good starting value, but it turns out that most of the tasks
         # relate to a particular pool in one way or another. Some tasks are given the poolname as a parameter,
         # and some can tell us by attaching a note to the message.
-        poolname: Optional[str] = None
+        poolname: ActorArgumentType = None
 
-        if 'poolname' in actor_arguments:
-            poolname = actor_arguments['poolname']
+        if 'poolname' in task_call.named_args:
+            poolname = task_call.named_args['poolname']
 
         elif message.options:
             poolname = get_metric_note(NOTE_POOLNAME)
@@ -364,3 +358,125 @@ class Prometheus(dramatiq.middleware.Middleware):  # type: ignore  # Class canno
             TaskMetrics.inc_overall_errored_messages(*labels)
 
     after_skip_message = after_process_message
+
+
+class WorkerMetrics(dramatiq.middleware.Middleware):  # type: ignore[misc]  # cannot subclass 'Middleware'
+    """
+    Dramatiq broker middleware spawning a thread to keep refreshing worker metrics.
+    """
+
+    def __init__(self, worker_name: str, interval: int) -> None:
+        super().__init__()
+
+        self.worker_name = worker_name
+        self.interval = interval
+
+        self._refresher: Optional[threading.Thread] = None
+
+    def after_worker_boot(self, signal: str, worker: dramatiq.worker.Worker) -> None:
+        from . import get_logger
+        from .metrics import WorkerMetrics as _WorkerMetrics
+
+        get_logger().warning('metrics refresher started')
+
+        self._refresher = _WorkerMetrics.spawn_metrics_refresher(
+            get_logger(),
+            self.worker_name,
+            self.interval,
+            lambda _worker: Ok((1, len(worker.workers))),
+            worker_instance=worker
+        )
+
+
+class WorkerTraffic(dramatiq.middleware.Middleware):  # type: ignore[misc]  # cannot subclass 'Middleware'
+    KEY_WORKER_TASK = 'tasks.workers.traffic.{worker}.{pid}.{tid}'  # noqa: FS003
+    KEY_WORKER_TASK_PATTERN = 'tasks.workers.traffic.*'
+
+    def __init__(
+        self,
+        logger: gluetool.log.ContextAdapter,
+        cache: redis.Redis,
+        worker_name: str
+    ) -> None:
+        super().__init__()
+
+        self.logger = logger
+        self.cache = cache
+        self.worker_name = worker_name
+        self.worker_pid = os.getpid()
+
+    @property
+    def current_key(self) -> str:
+        return self.KEY_WORKER_TASK.format(  # noqa: FS002
+            worker=self.worker_name,
+            pid=self.worker_pid,
+            tid=threading.get_ident()
+        )
+
+    def before_process_message(self, broker: dramatiq.broker.Broker, message: dramatiq.message.Message) -> None:
+        from .cache import set_cache_value
+        from .knobs import KNOB_WORKER_TRAFFIC_METRICS_TTL
+        from .metrics import WorkerTrafficTask
+        from .tasks import TaskCall
+
+        task_call = TaskCall.from_message(broker, message)
+
+        tid = threading.get_ident()
+
+        set_cache_value(
+            self.logger,
+            self.cache,
+            self.current_key,
+            WorkerTrafficTask(
+                workername=self.worker_name,
+                worker_pid=self.worker_pid,
+                worker_tid=tid,
+                ctime=datetime.datetime.utcnow(),
+                queue=cast(str, message.queue_name),
+                actor=cast(str, message.actor_name),
+                args=task_call.named_args
+            ).serialize_to_json().encode(),
+            ttl=KNOB_WORKER_TRAFFIC_METRICS_TTL.value
+        )
+
+    def after_process_message(
+        self,
+        broker: dramatiq.broker.Broker,
+        message: dramatiq.message.Message,
+        *,
+        # This is on purpose, our tasks never return anything useful.
+        result: None = None,
+        exception: Optional[BaseException] = None
+    ) -> None:
+        from .cache import delete_cache_value
+
+        delete_cache_value(self.logger, self.cache, self.current_key)
+
+    after_skip_message = after_process_message
+
+
+class CurrentMessage(dramatiq.middleware.Middleware):  # type: ignore[misc]  # cannot subclass 'Middleware'
+    """
+    Middleware that exposes the current message via a context variable.
+
+    Based on :py:class:`dramatiq.middleware.current_message.CurrentMessage`,
+    but modifies our context variables instead of managing its own storage.
+    """
+
+    def before_process_message(self, broker: dramatiq.broker.Broker, message: dramatiq.message.Message) -> None:
+        from .context import CURRENT_MESSAGE
+
+        CURRENT_MESSAGE.set(message)
+
+    def after_process_message(
+        self,
+        broker: dramatiq.broker.Broker,
+        message: dramatiq.message.Message,
+        *,
+        # This is on purpose, our tasks never return anything useful.
+        result: None = None,
+        exception: Optional[BaseException] = None
+    ) -> None:
+        from .context import CURRENT_MESSAGE
+
+        CURRENT_MESSAGE.set(None)
