@@ -1,18 +1,21 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import dataclasses
 import datetime
 import functools
-from typing import Callable, List, Tuple, Type, cast
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, cast
 
 import gluetool.log
 import gluetool.utils
 import sqlalchemy
 from gluetool.log import log_dict, log_table
 from gluetool.result import Error, Ok, Result
-from typing_extensions import Protocol
+from typing_extensions import Protocol, TypedDict
 
-from . import Failure, JSONType, partition
+from . import Failure, SerializableContainer, partition
 from .db import GuestRequest
-from .drivers import PoolCapabilities, PoolDriver
+from .drivers import PoolCapabilities, PoolDriver, PoolLogger
 from .knobs import Knob
 from .metrics import PoolMetrics
 
@@ -45,6 +48,49 @@ PolicyType = Callable[
 ]
 
 
+SerializedRulingHistoryItemType = TypedDict(
+    'SerializedRulingHistoryItemType',
+    {
+        'policyname': str,
+        'allowed-pools': List[str],
+        'cancel': bool,
+        'failure': Optional[Dict[str, Any]]
+    }
+)
+
+
+@dataclasses.dataclass
+class RulingHistoryItem(SerializableContainer):
+    policyname: str
+    ruling: Optional[PolicyRuling] = None
+    failure: Optional[Failure] = None
+
+    def serialize(self) -> SerializedRulingHistoryItemType:  # type: ignore[override]
+        serialized: SerializedRulingHistoryItemType = {
+            'policyname': self.policyname,
+            'allowed-pools': [],
+            'cancel': False,
+            'failure': None
+        }
+
+        if self.ruling is not None:
+            serialized['allowed-pools'] = [
+                pool.poolname
+                for pool in self.ruling.allowed_pools
+            ]
+
+            serialized['cancel'] = self.ruling.cancel
+
+        if self.failure is not None:
+            serialized['failure'] = self.failure.get_event_details()
+
+        return serialized
+
+    @classmethod
+    def unserialize(cls, serialized: SerializedRulingHistoryItemType) -> 'RulingHistoryItem':  # type: ignore[override]
+        raise NotImplementedError()
+
+
 class PolicyWrapperType(Protocol):
     policy_name: str
 
@@ -92,7 +138,11 @@ KNOB_ROUTE_POOL_ENABLED: Knob[bool] = Knob(
 
 class PolicyLogger(gluetool.log.ContextAdapter):
     def __init__(self, logger: gluetool.log.ContextAdapter, policy_name: str) -> None:
-        super(PolicyLogger, self).__init__(logger, {'ctx_policy_name': (50, policy_name)})
+        super().__init__(logger, {'ctx_policy_name': (50, policy_name)})
+
+    @property
+    def policyname(self) -> str:
+        return cast(str, self._contexts['policy_name'][1])
 
 
 def policy_boilerplate(fn: PolicyType) -> PolicyType:
@@ -195,7 +245,10 @@ def collect_pool_can_acquire(
     guest_request: GuestRequest
 ) -> Result[List[Tuple[PoolDriver, bool]], Failure]:
     r_answers = [
-        (pool, pool.can_acquire(logger, session, guest_request))
+        (
+            pool,
+            pool.can_acquire(PoolLogger(logger, pool.poolname), session, guest_request)
+        )
         for pool in pools
     ]
 
@@ -501,7 +554,7 @@ def policy_least_crowded(
 
     log_dict(logger.debug, 'pool metrics', pool_metrics)
 
-    min_usage = min([metrics.current_guest_request_count for _, metrics in pool_metrics])
+    min_usage = min(metrics.current_guest_request_count for _, metrics in pool_metrics)
 
     return Ok(PolicyRuling(
         allowed_pools=[
@@ -512,6 +565,83 @@ def policy_least_crowded(
     ))
 
 
+@policy_boilerplate
+def policy_most_free_addresses(
+    logger: gluetool.log.ContextAdapter,
+    session: sqlalchemy.orm.session.Session,
+    pools: List[PoolDriver],
+    guest_request: GuestRequest
+) -> PolicyReturnType:
+    """
+    Pick a pool with most free network addresses.
+    """
+
+    if len(pools) <= 1:
+        return Ok(PolicyRuling(allowed_pools=pools))
+
+    r_pool_metrics = collect_pool_metrics(pools)
+
+    if r_pool_metrics.is_error:
+        return Error(r_pool_metrics.unwrap_error())
+
+    pool_metrics = r_pool_metrics.unwrap()
+
+    # Hold usage report - pools and resources that were checked, and the outcome of the check.
+    usage_report: List[List[str]] = []
+
+    def pool_to_free_addresses(pool: PoolDriver, metrics: PoolMetrics) -> int:
+        free_addresses = 0
+
+        for network_name, network_limit in metrics.resources.limits.networks.items():
+            network_usage = metrics.resources.usage.networks.get(network_name)
+
+            # Networks that don't report any usage are treated as having enough resources - again, pool does not care
+            # about this network enough to provide data.
+            if network_usage is None or network_usage.addresses is None or network_limit.addresses is None:
+                network_free_addresses = 0
+
+            else:
+                network_free_addresses = network_limit.addresses - network_usage.addresses
+
+            free_addresses += network_free_addresses
+
+            usage_report.append([
+                pool.poolname,
+                'network.addresses',
+                str(network_limit.addresses) if network_limit else '',
+                str(network_usage.addresses) if network_usage else '',
+                str(network_free_addresses)
+            ])
+
+        return free_addresses
+
+    # Find out how many addresses each pool has...
+    pool_addresses = {
+        pool: pool_to_free_addresses(pool, metrics)
+        for pool, metrics in pool_metrics
+    }
+
+    # ... so we could get the biggest amount...
+    max_addresses = max(pool_addresses.values())
+
+    log_table(
+        logger.info,
+        'pool addresses',
+        [
+            ['Pool', 'Metric', 'Limit', 'Usage', 'Free']
+        ] + usage_report,
+        headers='firstrow',
+        tablefmt='psql'
+    )
+
+    # ... and allow all pools with this number of free addresses.
+    return Ok(PolicyRuling(allowed_pools=[
+        pool
+        for pool, addresses in pool_addresses.items()
+        if addresses == max_addresses
+    ]))
+
+
 @policy_boilerplate
 def policy_one_attempt_forgiving(
     logger: gluetool.log.ContextAdapter,
@@ -696,6 +826,29 @@ def policy_can_acquire(
     ))
 
 
+@policy_boilerplate
+def policy_use_only_when_addressed(
+    logger: gluetool.log.ContextAdapter,
+    session: sqlalchemy.orm.session.Session,
+    pools: List[PoolDriver],
+    guest_request: GuestRequest
+) -> PolicyReturnType:
+    """
+    Disallow pools that are marked as to be used only when requested by name.
+    """
+
+    if guest_request.environment.pool is not None:
+        return Ok(PolicyRuling(allowed_pools=pools))
+
+    return Ok(PolicyRuling(
+        allowed_pools=[
+            pool
+            for pool in pools
+            if pool.use_only_when_addressed is False
+        ]
+    ))
+
+
 def run_routing_policies(
     logger: gluetool.log.ContextAdapter,
     session: sqlalchemy.orm.session.Session,
@@ -723,18 +876,12 @@ def run_routing_policies(
     from .metrics import RoutingMetrics
 
     # Collecting all policy rulings along the way.
-    history: List[Tuple[str, PolicyRuling]] = []
+    history: List[RulingHistoryItem] = []
 
-    def _serialize_history() -> List[JSONType]:
-        # `dataclasses.asdict()`` not usable here, because `PolicyRuling.allowed_pools` contains `PoolDriver` instances
-        # which are not data classes.
+    def _serialize_history() -> List[SerializedRulingHistoryItemType]:
         return [
-            {
-                'policy': policy_name,
-                'allowed-pools': [pool.poolname for pool in policy_ruling.allowed_pools],
-                'cancel': policy_ruling.cancel
-            }
-            for policy_name, policy_ruling in history
+            history_item.serialize()
+            for history_item in history
         ]
 
     # Just a tiny helper, to avoid repeating the parameters...
@@ -757,15 +904,18 @@ def run_routing_policies(
         RoutingMetrics.inc_policy_called(policy_name)
 
         if r.is_error:
+            history.append(RulingHistoryItem(policyname=policy_name, failure=r.unwrap_error()))
+
             return Error(Failure.from_failure(
                 'failed to route guest request',
                 r.unwrap_error(),
+                environment=guest_request.environment,
                 history=_serialize_history()
             ))
 
         policy_ruling = r.unwrap()
 
-        history.append((policy_name, policy_ruling))
+        history.append(RulingHistoryItem(policyname=policy_name, ruling=policy_ruling))
 
         if policy_ruling.cancel:
             # Mark all input pools are excluded
