@@ -1,3 +1,6 @@
+# Copyright Contributors to the Testing Farm project.
+# SPDX-License-Identifier: Apache-2.0
+
 import imp
 import os
 from typing import Any, Callable, Dict
@@ -10,7 +13,7 @@ from . import Failure
 
 class ScriptEngine:
     def __init__(self) -> None:
-        super(ScriptEngine, self).__init__()
+        super().__init__()
 
         self.functions: Dict[str, Callable[..., Result[Any, Failure]]] = {}
         self.variables: Dict[str, Callable[..., Result[Any, Failure]]] = {}
